import functools
import importlib

from django.conf import (
    settings,
)

from behave_bo.loggers import (
    tests_logger,
)


class RunnerPluginsMixin:

    plugins = {}

    def get_plugins_packages(self):
        return [
            'behave_bo.behave_django_bo.plugins'
        ]

    def setup_plugins(self, behave_args, options):
        """Подключает плагины.

        Args:
            behave_args: Аргументы, относящиеся к behave_bo.
            options: Все аргументы команды запуска.

        """
        if options['plugins']:
            plugins_with_options = self._get_plugins_options(
                options['plugins'],
                options['plugin_vars'],
            )
            self._setup_plugins(
                plugins_with_options,
                options['disable_failed_plugin'],
            )

            for plugin_name, plugin_object in self.plugins.items():
                if hasattr(plugin_object, 'before_run'):
                    result = plugin_object.before_run()
                    behave_args.extend(result)

    def _setup_plugins(self, plugins_with_options, skip_if_failed=False):
        """Получает классы плагинов и инициализирует их с соответствующими параметрами.

        Args:
            plugins_options: dict. Словарь плагинов с параметрами.
            skip_if_failed: bool. Пропускать подключение при ошибке(не вызывать ошибку).

        """
        for plugin, options in plugins_with_options.items():
            import_errors = []
            plugin_class = None

            for package in self.get_plugins_packages():
                try:
                    plugin_module = importlib.import_module(f'{package}.{plugin}')
                    plugin_class = getattr(plugin_module, 'plugin_class')
                    break
                except ImportError as e:
                    import_errors.append(e)

            if plugin_class:
                try:
                    self.plugins[plugin] = plugin_class(**options)
                except Warning as w:
                    tests_logger.warning(w)
                except Exception as e:
                    if skip_if_failed:
                        tests_logger.warning(f"Can't import and setup plugin {plugin} due to {e}")
                    else:
                        tests_logger.fatal(f"Can't import and setup plugin {plugin} due to: {e}")
                        exit(-1)
            elif import_errors:
                if skip_if_failed:
                    tests_logger.warning(f"Can't import plugin '{plugin}':\n{import_errors}")
                else:
                    raise ImportError(f"Can't import plugin '{plugin}':\n{import_errors}")


    @staticmethod
    def _get_plugins_options(plugins, plugins_vars):
        """Получает опции плагинов из параметров командной строки.

        Args:
            plugins: str. Плагины.
            plugins_vars: Опции плагинов.

        Returns:
            dict. Словарь вида (наименование плагина: опции плагина).

        """
        plugin_list = plugins.split(',')
        plugins_dict = {name: {} for name in plugin_list}

        if plugins_vars:
            plugin_vars_list = plugins_vars.split(',')
            for plugin in plugin_list:
                for plugin_var in plugin_vars_list:
                    if plugin_var.startswith(f'{plugin}_'):
                        plugin_var = plugin_var.removeprefix(f'{plugin}_')
                        try:
                            key, value = plugin_var.split('=')
                        except ValueError:
                            key, value = plugin_var, True
                        plugins_dict[plugin][key] = value

        return plugins_dict


class PluginException(Exception):
    pass


class PluginWarning(Warning):
    pass


class Plugin:

    REQUIRED_SETTINGS = []

    def before_all(self, context):
        pass

    def before_feature(self, context, feature):
        pass

    def before_scenario(self, context, scenario):
        pass

    def before_step(self, context, step):
        pass

    def after_step(self, context, step):
        pass

    def after_scenario(self, context, scenario):
        pass

    def after_feature(self, context, feature):
        pass

    def after_all(self, context):
        pass

    def check_required_settings(self):
        """Проверяет необходимые настройки для плагина.

        Raises:
            PluginException: при отсутствии настроек self.REQUIRED_SETTINGS.

        """
        missing_settings = []

        for setting_name in self.REQUIRED_SETTINGS:
            if not hasattr(settings, setting_name):
                missing_settings.append(setting_name)

        if missing_settings:
            raise PluginException(
                f'{self.__class__.__name__} requires settings: {", ".join(missing_settings)}.'
            )


def apply_plugins(plugins_dict, hook_name, *args, **kwargs):
    for plugin_name, plugin_object in plugins_dict.items():
        try:
            plugin_action = getattr(plugin_object, hook_name)
            plugin_action(*args, **kwargs)
        except AssertionError as e:
            raise e
        except Exception as e:
            tests_logger.warning(f'Plugin action "{plugin_name}" failed due to {e}')


def use_plugins(hook):
    """Декоратор для запуска hook'ов в классе плагина.

    """
    @functools.wraps(hook)
    def wrapper(self, *args, **kwargs):
        context = args[0]
        plugins = context.config.django_test_runner.plugins

        if hook.__name__.startswith('before_'):
            # Если это before-хук, сначала выполняем код из плагинов, а после него код хука.
            apply_plugins(plugins, hook.__name__, *args, **kwargs)

        result = hook(self, *args, **kwargs)

        if hook.__name__.startswith('after_'):
            # Если это after-хук, сначала выполняем код хука, а после него код из плагинов.
            apply_plugins(plugins, hook.__name__, *args, **kwargs)

        return result

    return wrapper



import logging
import os
import uuid

from behave_bo.loggers import (
    tests_logger,
)
from django.conf import (
    settings,
)


class FixChecksLogger:
    """Контекстный менеджер, выполняющий настройку логирования в файл при исправлении данных в проверках.

    Attributes:
        enable: Булевый флаг. Если True - будет выполнена настройка логгера и вывод в консоль при выходе.
        log_file_path: Путь к лог-файлу.
    """
    logger_name = 'fix_checks_logger'

    def __init__(self, enable: bool):
        """Инициализация FixChecksLogger."""
        self.enable = enable
        self.log_file_path = (
            os.path.join(settings.DOWNLOADS_DIR, f'fix_checks_{str(uuid.uuid4())[:8]}.txt')
        )

    def __enter__(self):
        if self.enable:
            self._setup_logger()
            tests_logger.info(f'Исправления по тегу --fix-checks будут сохранены в {self.log_file_path}.\n')

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.enable:
            if os.path.exists(self.log_file_path):
                message = (
                    f'Файл с изменениями по тегу --fix-checks сохранён в {self.log_file_path}\n'
                    f'Данный файл необходимо приложить в PR задачи!!!\n'
                )
            else:
                message = 'Исправлений по тегу --fix-checks не внесено. Файл не создан.\n'

            tests_logger.info(message)

    def _setup_logger(self):
        """Выполняет настройку логгера."""
        log_setup = logging.getLogger(self.logger_name)
        formatter = logging.Formatter(
            fmt='%(levelname)s: %(asctime)s %(message)s',
            datefmt='%m/%d/%Y %I:%M:%S %p',
        )
        filehandler = logging.FileHandler(
            self.log_file_path,
            mode='a',
            delay=True,
        )
        filehandler.setFormatter(formatter)
        log_setup.setLevel(logging.ERROR)
        log_setup.addHandler(filehandler)


fix_checks_logger = logging.getLogger(FixChecksLogger.logger_name)



import datetime

import time_machine

from django.conf import (
    settings,
)
from django.test.runner import (
    DiscoverRunner,
)

from behave_bo.__main__ import (
    run_behave,
)
from behave_bo.configuration import (
    Configuration,
)
from behave_bo.loggers import (
    tests_logger,
)
from behave_bo.reporter.failed_tags import (
    FailedTagsParallelReporter,
    FailedTagsReporter,
)
from behave_bo.reporter.junit import (
    JUnitReporter,
)
from behave_bo.runner import (
    BehaveRunner,
    BehaveTagsCachedRunner,
    ParallelBehaveRunner,
)
from behave_bo.tag_expression import (
    TagExpression,
)


class BaseBehaveTestRunner(DiscoverRunner):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.return_code_success = 0
        self.behave_config = None

        if kwargs.get('save_db'):
            # В случае если нужно сохранить результат в БД / отдельную БД,
            # атрибут keepdb должен быть True
            self.keepdb = True

    def get_tempdir_path(self):
        raise NotImplementedError

    def get_template_data_types(self):
        raise NotImplementedError

    def get_environment_filepath(self):
        raise NotImplementedError

    def run_behave_tests(self, behave_args, options):
        """Выполняет запуск behave тестов.

        Args:
            behave_args: Аргументы, относящиеся к behave_bo.
            options: Все аргументы команды запуска.

        Returns:
            Результат выполнения тестов.

        """
        self.behave_config = Configuration(behave_args)
        try:
            log_capture_exclude_template = settings.WEB_BB_BEHAVE__LOG_CAPTURE_EXCLUDE_TEMPLATE
        except AttributeError:
            log_capture_exclude_template = ''

        self.behave_config.log_capture_exclude_template = log_capture_exclude_template

        self._add_extra_parameters_to_behave_config(options)

        traveller = None

        if self.behave_config.replace_current_date:
            traveller = time_machine.travel(self.behave_config.replace_current_date)
            traveller.start()

        if self.parallel:
            behave_runner_class = ParallelBehaveRunner
            failed_tags_reporter_class = FailedTagsParallelReporter
        else:
            if self.behave_config.optimize_steps_loading:
                behave_runner_class = BehaveTagsCachedRunner
            else:
                behave_runner_class = BehaveRunner
            failed_tags_reporter_class = FailedTagsReporter

        failed_tags_reporter = failed_tags_reporter_class(self.behave_config)
        self.behave_config.reporters.append(failed_tags_reporter)

        try:
            junit_reporter = [r for r in self.behave_config.reporters if isinstance(r, JUnitReporter)][0]
        except IndexError:
            junit_reporter = None

        runner_kwargs = {
            'tempdir': self.get_tempdir_path(),
            'step_template_data_types': self.get_template_data_types(),
            'environment_filepath': self.get_environment_filepath(),
            'test_runner': self,
        }

        run_result = run_behave(
            self.behave_config,
            runner_class=behave_runner_class,
            runner_kwargs=runner_kwargs,
        )

        while failed_tags_reporter.rerun_required():
            failed_tags_reporter.rerun_up()
            tests_logger.info(
                f'Перезапуск упавших сценариев для выявления ложных падений '
                f'среди {len(failed_tags_reporter.main_run_failed_tags)} сценариев: '
                f'{", ".join(failed_tags_reporter.main_run_failed_tags)} '
                f'(попытка №{failed_tags_reporter.rerun_attempt})...\n'
            )

            if junit_reporter:
                junit_reporter.rerun = True
                self.behave_config.reporters.append(junit_reporter)

            run_behave(
                self.behave_config,
                runner_class=behave_runner_class,
                runner_kwargs=runner_kwargs,
            )

            if failed_tags_reporter.check_all_false_positive():
                run_result = self.return_code_success

        if traveller:
            traveller.stop()

        return run_result

    def _add_extra_parameters_to_behave_config(self, options):
        """Добавляет дополнительные параметры в конфиг в соответствии с options.

        Args:
            options: Аргументы команды запуска.

        """
        config = self.behave_config
        config.django_test_runner = self

        config.parallel = options.get('parallel')
        config.parallel_count = options.get('parallel_count')
        config.collect_top_scenarios = options.get('collect_top_scenarios')
        config.save_db = options.get('save_db')
        config.optimize_steps_loading = options.get('optimize_steps_loading')
        config.clear_cached_step_locations = options.get('clear_cached_step_locations')
        config.optimize_features_steps_loading = options.get('optimize_features_steps_loading')
        config.in_tags_order = options.get('in_tags_order')
        config.rerun_if_failed = options.get('rerun_if_failed')
        config.rerun_attempts = options.get('rerun_attempts')
        config.measure_requests_coverage = options.get('measure_requests_coverage')
        config.replace_current_date = None

        if options.get('replace_current_date'):
            try:
                config.replace_current_date = datetime.datetime.strptime(options['replace_current_date'], '%d.%m.%Y')
                tests_logger.info(
                    f'Запуск автотестов с заменой текущей даты на {config.replace_current_date:%d.%m.%Y}'
                )
            except ValueError as e:
                tests_logger.warning(f'Некорректное значение для параметра replace_current_date: {e}')

        config.save_changes_to_db = bool(config.save_db)

        if options.get('from_file_tags'):
            file_path = options.get('from_file_tags')
            try:
                with open(file_path) as f:
                    tags_from_file = f.readlines()[0]
                    config.tags = TagExpression([tags_from_file])
                    tests_logger.warning(
                        f'! Запуск автотестов по тегам перечисленным в файле {file_path}: {tags_from_file}'
                    )
            except Exception as e:
                tests_logger.warning(
                    f'Не удалось получить данные из файла с перечислением тегов для запуска автотестов {file_path}: {e}'
                )

        return config



from django.dispatch import (
    Signal,
)


skip_request_check_signal = Signal()
after_step_plugins = Signal(providing_args=['context', 'step'])
after_scenario_signal = Signal(providing_args=['context', 'scenario'])



import datetime
import json
import os
import shutil
import tempfile
import traceback
import uuid
from collections import (
    defaultdict,
)
from contextlib import (
    contextmanager,
)
from itertools import (
    zip_longest,
)

import time_machine
from django.apps import (
    apps,
)
from django.conf import (
    settings,
)
from django.db.models.signals import (
    post_delete,
    post_init,
    post_save,
)
from django.test.utils import (
    override_settings,
)

from behave_bo.loggers import (
    tests_logger,
)


def get_features_paths():
    """Возвращает список путей до директорий features в app'ах, если существуют."""

    paths = []
    for app in apps.get_app_configs():
        features_dir = os.path.join(app.path, 'features')
        if os.path.isdir(features_dir):
            paths.append(features_dir)

    return paths


@contextmanager
def override_dirs(scenario):
    """Контекстный менеджер для создания уникальных директорий под временные файлы и downloads для выполняемого сценария

    Args:
        scenario: объект класса Scenario

    """
    prefix = scenario.main_tag if scenario.tags else uuid.uuid4().hex[:6]

    old_tempdir = settings.TEMPDIR
    old_downloads_dir = settings.DOWNLOADS_DIR

    new_tempdir = f"{old_tempdir}/{prefix}__{scenario.line}"
    new_downloads_dir = f"{old_downloads_dir}/{prefix}__{scenario.line}"

    if os.path.isdir(new_tempdir):
        shutil.rmtree(new_tempdir, ignore_errors=True)

    tempfile.tempdir = new_tempdir

    if not os.path.isdir(tempfile.gettempdir()):
        os.makedirs(tempfile.gettempdir())

    if os.path.isdir(new_downloads_dir):
        shutil.rmtree(new_downloads_dir, ignore_errors=True)

    if not os.path.isdir(new_downloads_dir):
        os.makedirs(new_downloads_dir)

    try:
        with override_settings(TEMPDIR=new_tempdir, DOWNLOADS_DIR=new_downloads_dir):
            yield
    finally:
        tempfile.tempdir = old_tempdir


@contextmanager
def override_current_date(selected_date):
    """Контекстный менеджер для подмены текущей даты

    Args:
        selected_date: дата на которую требуется подменить текущую дату
    """
    try:
        if selected_date:
            with time_machine.travel(selected_date):
                yield
        else:
            yield
    finally:
        pass


class ModelRecordAnalyzer:
    """Инструмент для анализа очередности создания, удаления и инициализации записей моделей."""

    class Aside:
        """Класс для хранения промежуточного состояния."""
        def __init__(self):
            self.prev_step_id = None
            self.step_id = None
            self.model_label = None
            self.pk_list = []
            self.action = None
            self.stack = None
            self.hash_on_stack = 0

    def __init__(self, context, *args, **kwargs):
        self.context = context

        row_id = ''

        if context.scenario._row:
            row_id = f'_{context.scenario._row.id}'

        self.scenario_id = f'{context.scenario.main_tag}{row_id}'

        self.step_log = {self._get_step_id(step): [] for step in context.scenario.all_steps}
        self.stack_log = defaultdict()
        self._connect()

        self._counter = 0
        self.aside = self.Aside()

    def _connect(self):
        """Подключение к сигналам django."""
        post_save.connect(self._save_post_save)
        post_delete.connect(self._save_post_delete)
        post_init.connect(self._save_post_init)

    def _save_post_save(self, sender, instance, created, **kwargs):
        """Обработчик post_save сигнала."""
        self._lazy_save(sender, instance.pk, 'created' if created else 'updated')

    def _save_post_delete(self, sender, instance, **kwargs):
        """Обработчик post_delete сигнала."""
        self._lazy_save(sender, instance.pk, 'deleted')

    def _save_post_init(self, sender, instance, **kwargs):
        """Обработчик post_init сигнала."""
        if instance.pk and not self._called_from_factory():
            self._lazy_save(sender, instance.pk, 'evaluated')

    def _called_from_factory(self):
        """Выполняется ли вызов через Factory."""
        # TODO BOBUH-25284 web_bb_behave/runner путь из проекта убрать
        return any('web_bb_behave/runner/factories/base.py' in p.filename for p in self._get_stack())

    @property
    def _dir_path_to_save(self):
        """Директория хранения логов по выполненным тестам."""
        dir_to_save = os.path.join(settings.FACTORY_BO__FIXTURES_DIR_PATH, 'testing_analyze_model_records')
        if not os.path.exists(dir_to_save):
            os.mkdir(dir_to_save)

        return dir_to_save

    @property
    def _file_path_to_save(self):
        """Путь к файлу для сохранения результата."""
        timestamp = f'{datetime.datetime.now():%d%m%y%H%M%S}'
        file_path_to_save = os.path.join(self._dir_path_to_save, f'{self.scenario_id}_{timestamp}')

        return file_path_to_save

    @property
    def _step_id(self):
        """Идентификатор шага."""
        if hasattr(self.context, 'current_step'):

            return self._get_step_id(self.context.current_step)

    @staticmethod
    def _get_step_id(step):
        """Поучает идентификатор шага."""

        number = getattr(step, 'number', 0)

        return f'{number}. {step.keyword} {step.name}'

    def _next_num(self):
        """Увеличивает счетчик на 1 и возвращает полученное значение."""
        self._counter += 1

        return self._counter

    @staticmethod
    def _in_projects_dirs(path):
        """Проверяет, относится ли путь к основным директориям проекта."""

        return any(repo in path for repo in settings.WEB_BB_BEHAVE__TESTING_PROJECTS)

    def _get_stack(self):
        """Получает стек по директориям проекта."""
        stack = [s for s in traceback.extract_stack() if self._in_projects_dirs(s.filename)]

        return stack

    def _get_hash_on_stack(self):
        """Получает хэш по стеку, без учета последних 3 строк, относящихся к вызовам внутри данного инструмента."""
        stack = [f'{s.filename}:{s.lineno}' for s in self._get_stack()][:-3]

        return hash(''.join(stack))

    def _put_aside(self, sender, pk, action, hash_on_stack, reset_pk=False):
        """Запоминает состояние вызова."""
        if reset_pk:
            self.aside.pk_list.clear()

        self.aside.step_id = self._step_id
        self.aside.model_label = sender._meta.label
        self.aside.action = action
        self.aside.pk_list.append(pk)
        self.aside.hash_on_stack = hash_on_stack
        self.aside.stack = self._get_stack()

    def _save(self):
        """Сохраняет состояние aside в основной лог выполнения."""
        if self.aside.prev_step_id != self.aside.step_id:
            self.aside.prev_step_id = self.aside.step_id
            self._counter = 0
        number = self._next_num()
        string_to_save = f'{number} {self.aside.model_label} {self.aside.pk_list} {self.aside.action}'

        self.step_log[f'{self.aside.step_id}'].append(string_to_save)
        self.stack_log[string_to_save] = self.aside.stack

    def _lazy_save(self, sender, pk, action):
        """Отложенное сохранение результата выполнения."""
        if self._step_id:
            hash_on_stack = self._get_hash_on_stack()
            # При первом вызове сохраняет состояние в aside.
            if not self.aside.hash_on_stack:
                self._put_aside(sender, pk, action, hash_on_stack)
            # При совпадении стека, модели и действия расширяет список aside.pk_list.
            # Таким образом выполняется отбор вызовов относящихся к одному queryset.
            elif (
                hash_on_stack == self.aside.hash_on_stack
                and self.aside.model_label == sender._meta.label
                and self.aside.action == action
            ):
                self._put_aside(sender, pk, action, hash_on_stack)
            # Иначе aside.pk_list сбрасывается.
            else:
                self._save()
                self._put_aside(sender, pk, action, hash_on_stack, reset_pk=True)

    def _parse_log_file(self, file_path):
        """Парсит лог файл."""
        with open(file_path) as f:
            json_log = json.loads(f.read())

        return json_log

    def _get_last_created_log_file_for_scenario(self):
        """Получает путь к последнему созданному файлу для сценария."""
        last_created_log_file = None

        for file_name in sorted(os.listdir(self._dir_path_to_save), reverse=True):
            if file_name.startswith(f'{self.scenario_id}'):
                last_created_log_file = os.path.join(self._dir_path_to_save, file_name)
                break

        return last_created_log_file

    def _get_previous_log_for_scenario(self):
        """Получает лог последнего запуска для сценария."""
        log = None

        last_created_log_file = self._get_last_created_log_file_for_scenario()
        if last_created_log_file:
            log = self._parse_log_file(last_created_log_file)

        return log

    def _compare_with_previous_log(self):
        """Выполняет сравнение с результатом предыдущего запуска, если такой был."""
        prev_log = self._get_previous_log_for_scenario()
        if not prev_log:
            tests_logger.info('Previous log file not found.')
        elif prev_log.keys() != self.step_log.keys():
            tests_logger.info('Steps don\'t match.')
        else:
            for step_name, prev_list, curr_list in zip(prev_log.keys(), prev_log.values(), self.step_log.values()):
                for prev, curr in zip_longest(prev_list, curr_list, fillvalue=None):
                    if prev != curr:
                        try:
                            stack = self.stack_log[curr]
                        except KeyError:
                            stack = 'Failed to get traceback.'
                        else:
                            stack = ''.join(traceback.format_list(stack))
                        tests_logger.warning(
                            f'First difference found in step {step_name}\n'
                            f'Previous: {prev}\n'
                            f'Current:  {curr}\n'
                            f'Traceback:\n'
                            f'{stack}'
                        )
                        break
                else:
                    continue
                break

    def write(self):
        """Запись результата в файл."""
        self._save()
        self._compare_with_previous_log()

        with open(self._file_path_to_save, 'w') as f:
            string_to_write = json.dumps(self.step_log, indent=2, ensure_ascii=False)
            f.write(string_to_write)



import ast
import fileinput
import json
import os
import re
import shutil
import traceback
from datetime import (
    datetime,
)
from pprint import (
    pformat,
)

from django.conf import (
    settings,
)
from django.test.testcases import (
    TestCase,
    TransactionTestCase,
)

from behave_bo.behave_django_bo.helpers import (
    context_extractor,
    filter_dict_keys,
    sort_lists_in_dict,
)
from behave_bo.behave_django_bo.loggers import (
    fix_checks_logger,
)
from behave_bo.loggers import (
    tests_logger,
)

from behave_bo.behave_django_bo.helpers import (
    FilesCompare,
)


class ReplaceChecksDataMixin:
    """Миксин добавляющий в класс тесткейса метод перезаписи данных для проверок."""

    def replace_checks_data(self, prev_frame, raw):
        """
        Заменить данные в автоматически сгенерированных проверка.

        Args:
            prev_frame: объект класса traceback.StackSummary необходимый для доступа к файлу проверок
                и номеру строки на которой выполнялся вызов метода проверки.
            raw: Данные на которые требуется произвести замену.

        """
        line_no = prev_frame.lineno + 1  # Следующая строка т.к. сами данные для проверки указываются на новой строке.

        for idx, line in enumerate(
            fileinput.input([prev_frame.filename], inplace=True),
            start=1,
        ):
            if line_no == idx:
                if re.search(': (f\'|f"|context)', line):
                    params = re.findall("(\'[a-zA-Z0-9_]+\': .*?),", line)
                    params_from_context = [param for param in params if 'context' in param]
                    error_msg = (
                        f'Проверочные данные в строке {prev_frame.lineno + 1} '
                        f'содержат переменные {params_from_context} '
                        f'и не могут быть полностью заменены данными ответа на запрос в приложение.'
                    )
                    tests_logger.error(error_msg)
                    fix_checks_logger.error(error_msg)

                    print(line, end='')
                else:
                    msg_splitter = ", msg='"
                    line = line.strip()
                    if msg_splitter in line:
                        line, end_of_line = line.split(msg_splitter)
                        end_of_line = f'{msg_splitter}{end_of_line}'
                    else:
                        end_of_line = ','

                    actual_part = ast.get_source_segment(
                        line,
                        ast.parse(line).body[0].value.elts[0],
                    )
                    new_expected_part = pformat(json.loads(raw), width=1000000)

                    new_line = f'{" " * 8}{actual_part}, {new_expected_part}{end_of_line}'
                    print(f'{new_line}')
            else:
                print(line, end='')


class AddExcludedKeysMixin:
    """Миксин добавляющий в класс тесткейса метод добавления (или дополнения существующего) excluded_keys."""

    def __init__(self):
        self.counter = 0 # счетчик количества добавленных строк в файле проверок
        super().__init__()

    def add_excluded_keys(self, prev_frame, params_to_add_to_excluded_keys):
        """
        Добавить новый или дополнить старый excluded_keys.

        Args:
            prev_frame: объект класса traceback.StackSummary необходимый для доступа к файлу проверок
                и номеру строки на которой выполнялся вызов метода проверки.
            params_to_add_to_excluded_keys: Данные на которые требуется добавить в excluded_keys.

        """
        # Следующая строка с учётом ранее добавленных строк, т.к. исключения для проверки указываются на новой строке.
        line_no = prev_frame.lineno + self.counter + 2

        for idx, line in enumerate(
            fileinput.input([prev_frame.filename], inplace=True),
            start=1,
        ):
            if line_no == idx:
                if 'excluded_keys' in line:
                    old_excluded_keys = re.search(r'excluded_keys=[\(|\[](.*?),?[\)|\]]', line).group(1)
                    if old_excluded_keys == '':
                        new_excluded_keys = params_to_add_to_excluded_keys
                    else:
                        new_excluded_keys = (
                                tuple(old_excluded_keys.replace('\'', '').replace(' ', '').split(',')) +
                                params_to_add_to_excluded_keys
                        )
                    new_line = f'{" " * 8}excluded_keys={tuple(set(list(new_excluded_keys)))},'
                else:
                    line = line.replace('\n', '')
                    new_line = f'{" " * 8}excluded_keys={params_to_add_to_excluded_keys},\n{line}'
                    self.counter += 1
                print(f'{new_line}')
            else:
                print(line, end='')


class CheckJsonWithExclusionMixin(ReplaceChecksDataMixin, AddExcludedKeysMixin):
    """Миксин добавляющий в класс тесткейса метод проверки JSON с исключением ключей."""

    files_compare_class = FilesCompare
    assertJSONEqual: callable
    runner = None
    common_excluded_keys = {
        'id',
        'pk',
    }

    def _get_content_from_current_response(self):
        """Получает содержимое текущего ответа на запрос."""
        check_func_stack_level = -3  # уровень вызова функции сравнения данных, относительно текущего уровня
        response_name = re.search(
            r'\(context\.(.*)\.context',
            traceback.extract_stack()[check_func_stack_level].line
        ).group(1)
        response = getattr(self.runner.behave_runner.context, response_name)

        try:
            response_content = json.loads(response.content)
        except json.JSONDecodeError:
            response_content = response.content.decode()

        return response_content

    def get_context_json(self, context):
        """Извлекает из контекста объект компонента и преобразует его в json-строку.

        Args:
            context: Объект списка контекстов ContextList

        Returns:
            JSON-строка.
        """
        assert context is not None, (
            f'Не удалось получить данные окна из объекта контекста. Вместо этого получен ответ:\n'
            f'{self._get_content_from_current_response()}'
        )

        component = None

        if 'window' in context:
            component = context['window']
        elif 'component' in context:
            component = context['component']

        return context_extractor.prepare_component_json(component) if component else '{}'

    def get_content_file(self, content) -> str:
        """Формирует абсолютный путь до скаченного файла

        Args:
            content: Контент ответа на запрос

        Returns:
            Строка с абсолютным путем до файла.
        """

        file_name = context_extractor.parse_download_file_name(
            content.decode('utf-8'),
            settings.DOWNLOADS_URL,
        )

        return os.path.join(settings.DOWNLOADS_DIR, file_name)

    def get_content_data_with_regex(self, content) -> str:
        """Получает из js-ответа данные с помощью регулярных выражений.

        Args:
            content: Контент ответа на запрос.

        Returns:
            JSON-строка.
        """
        content_data = context_extractor.extract_data_with_regex(content)
        if content_data:
            result = json.dumps(content_data)
        else:
            raise Exception('Regex found nothing')

        return result

    def assertFileEqual(self, context, file_download, file_compare, compare_file_names=True,
                        file_name_mask=None, current_date_in_file_name=False):
        """Проверяет утверждение что файлы совпадают

        Args:
            context: Объект контекста выполнения автотеста behave_bo.runner.Context
            file_download: Абсолютный путь до файла.
            file_compare: Относительный путь до файла.
            compare_file_names: Проверять или нет названия файлов на совпадение
            file_name_mask: Маска названия загружаемого файла
            current_date_in_file_name: Присутствует ли текущая дата в названии выгружаемого файла

        Examples:
            file_name_mask: r'Реестр_Шаблоны_видов_оплат_и_удержаний_15.07.2024_[0-9a-f]{16}.xls'

        Raises:
            AssertionError: в случае выявления различий сравниваемых файлов
        """

        def check_and_replace_current_date(file_name_mask):
            """
            Метод проверки и замены текущей даты в названии файла

            Args:
                file_name_mask: Маска названия загружаемого файла

            Returns:
                 file_name_mask: Маска названия загружаемого файла с заменой найденной даты на текущую в том же формате
            """
            now = datetime.now()
            current_date_formats = [
                "%d\\.%m\\.%Y",
                "%Y%m%d",
                "%Y\\.%m\\.%d",
            ]
            current_date_strings = [now.strftime(el) for el in current_date_formats]

            date_patterns = [
                '_?(\\d{2}\\\\.\\d{2}\\\\.20\\d{2})_',
                r'_?(20\d{2}\d{2}\d{2})_',
                '_?(20\\d{2}\\\\.\\d{2}\\\\.\\d{2})_',
            ]

            for date_str, pattern in zip(current_date_strings, date_patterns):
                matches = re.findall(pattern, file_name_mask)
                if matches:
                    file_name_mask = file_name_mask.replace(matches[0], date_str)
                    break

            return file_name_mask

        if compare_file_names:
            file_compare_name = os.path.basename(file_compare)
            file_download_name = str(file_download).split(f'{settings.DOWNLOADS_DIR}/')[1].replace('/', '_')

            if file_name_mask:
                mask_type = 'Заданный шаблон'
                if current_date_in_file_name:
                    file_name_mask = check_and_replace_current_date(file_name_mask)
                file_names_match = bool(re.match(file_name_mask, file_download_name))
            else:
                mask_type = 'Созданный шаблон'
                file_names_match = file_compare_name == file_download_name

                if not file_names_match:
                    file_compare_name = file_compare_name.replace('[', r'\[').replace(']', r'\]')
                    # Формирование маски по имени эталонного файла.
                    for hex_uuid_len in range(32, 7, -4):
                        re_hex_uuid = f'[0-9a-f]{{{hex_uuid_len}}}'
                        uuid_match = re.match(f'(.*)?({re_hex_uuid})(.*)?', file_compare_name)

                        if uuid_match:
                            file_name_mask = ''.join((uuid_match.group(1), re_hex_uuid, uuid_match.group(3)))
                            file_name_mask = file_name_mask.replace(
                                '.', r'\.').replace('(', r'\(').replace(')', r'\)')
                            if current_date_in_file_name:
                                file_name_mask = check_and_replace_current_date(file_name_mask)

                            file_names_match = bool(re.match(file_name_mask, file_download_name))
                            break
                    else:
                        file_name_mask = 'Не удалось сформировать шаблон.'

            if not file_names_match:
                error_message = (
                    f'Название полученного файла не совпадает с эталонным или не соответствуют шаблону!\n'
                    f'Полученный файл:\t{file_download_name}\n'
                    f'Эталонный файл:\t{file_compare_name}\n'
                    f'{mask_type}:\t{file_name_mask}\n'
                )
                raise AssertionError(error_message)

        compare = self.files_compare_class(table=context.table)
        _, extension = os.path.splitext(file_download)
        file_compare_path = os.path.join(context.scenario.location.dirname(), file_compare)

        if extension[1:].upper() == 'ZIP':
            result_compare = compare.compare_zip_archives(
                file_download, file_compare_path
            )
        else:
            result_compare = compare.compare_by_type(
                file_download, file_compare_path
            )

        if not result_compare:
            format_frame = traceback.format_stack()[-2]
            error_message = (
                f'{format_frame.splitlines()[0]}\n'
                f'Файл:\n'
                f'\t{file_download}\n'
                f'не совпадает по содержанию c эталонным файлом:\n'
                f'\t{file_compare_path}!\n'
                f'{compare.message}'
            )

            if self.runner.behave_config.fix_checks:
                os.remove(file_compare_path)
                shutil.copy(file_download, file_compare_path)
                tests_logger.error(error_message)
                fix_checks_logger.error(error_message)
            else:
                raise AssertionError(error_message)

    def get_sort_key(self, item):
        """Получить значения ключа для сортировки."""
        if isinstance(item, dict):
            item_list = [(key, self.get_sort_key(value)) for key, value in item.items()]
            key = sorted(item_list, key=lambda x: (x[1] is None, x))
        elif isinstance(item, list):
            item.sort(key=self.get_sort_key)
            key = item
        else:
            key = str(item)

        return key

    def check_sorted_json_equal(
        self,
        fact_data,
        expected_data,
    ) -> bool:
        """Проверяет на равенство списки и словари с отсортированными списками.

        Args:
            fact_data: Проверяемые данные.
            expected_data: Ожидаемые данные.

        Returns:
            Результат сравнения.
        """
        sorted_fact_data = sort_lists_in_dict(fact_data, sort_key=self.get_sort_key)
        sorted_expected_data = sort_lists_in_dict(expected_data, sort_key=self.get_sort_key)

        result = True
        try:
            self.assertJSONEqual(
                json.dumps(sorted_fact_data),
                sorted_expected_data,
            )
        except AssertionError:
            result = False

        return result

    def assertJSONEqualWithExclusion(
        self,
        raw: str,
        expected_data: dict,
        excluded_keys=None,
        non_excluded_keys=None,
        excluded_grid_indexes=None,
        exclude_sorting=False,
        msg=None,
    ):
        """Проверяет утверждение что сырой JSON соответствует ожидаемому словарю.

        Только при условии исключения значений соответствующих ключам
        из self.common_excluded_keys и excluded_keys
        и включения значение соответствующих ключам из non_excluded_keys.

        Args:
            raw: JSON-строка.
            expected_data: Словарь который ожидаем получить.
            excluded_keys: Список ключей для исключения.
            non_excluded_keys: Список ключей для не исключения.
            excluded_grid_indexes: Список индексов для исключения из проверки по ключу grid.
            exclude_sorting: Исключить проверку соответствия сортировки записей.
            msg: Сообщение выводимое при не прохождении проверки.
        """
        if not excluded_keys:
            excluded_keys = set()
        if not non_excluded_keys:
            non_excluded_keys = set()

        all_excluded_keys = self.common_excluded_keys.copy()
        all_excluded_keys.update(set(excluded_keys))
        all_excluded_keys = all_excluded_keys.difference(set(non_excluded_keys))

        filtered_raw = filter_dict_keys(
            json.loads(raw),
            all_excluded_keys,
        )

        expected_data = filter_dict_keys(
            expected_data,
            all_excluded_keys,
        )

        if excluded_grid_indexes:
            for index in sorted(excluded_grid_indexes, reverse=True):
                [el.pop(index) for el in filtered_raw['grid']]
                [el.pop(index) for el in expected_data['grid']]

        if exclude_sorting and filtered_raw and expected_data:
            filtered_raw = sort_lists_in_dict(filtered_raw, sort_key=self.get_sort_key)
            expected_data = sort_lists_in_dict(expected_data, sort_key=self.get_sort_key)

        error_msg = 'Данные из response не соответствуют ожидаемым!'

        if msg:
            error_msg = f'{error_msg} {msg}'

        try:
            self.assertJSONEqual(
                json.dumps(filtered_raw),
                expected_data,
                error_msg,
            )
        except Exception as exc:
            if not exclude_sorting and filtered_raw and expected_data:
                sorted_json_are_equal = False

                try:
                    sorted_json_are_equal = self.check_sorted_json_equal(
                        filtered_raw,
                        expected_data,
                    )
                except Exception as e:
                    tests_logger.error(f'Невозможно отсортировать записи для проверки различия сортировки: {e}')

                if sorted_json_are_equal:
                    error_msg += ' ОТЛИЧАЕТСЯ СОРТИРОВКА!'

            format_frame = traceback.format_stack()[-2]
            tests_logger.error(f'{error_msg}\n{format_frame.splitlines()[0]}')

            if self.runner.behave_config.fix_checks:
                exc_lines = "\n".join(
                    line for line in str(exc).split('\n')
                    if any(line.strip().startswith(char) for char in ('?', '+', '-'))
                )

                tests_logger.error(f'\n{exc_lines}')
                fix_checks_logger.error(f'{error_msg}\n{format_frame.splitlines()[0]}\n{exc_lines}\n')

                prev_frame = traceback.extract_stack()[-2]
                self.replace_checks_data(prev_frame, raw)
            elif self.runner.behave_config.fix_excluded_keys:
                params_to_add_to_excluded_keys = tuple(
                    key for key in self.runner.behave_config.fix_excluded_keys.split(',')
                )
                tests_logger.error(f'\nКлючи для добавления в excluded_keys: {params_to_add_to_excluded_keys}')

                fix_checks_logger.error(f'fix_excluded_keys: {error_msg}\n{format_frame.splitlines()[0]}\n'
                                        f'Ключи для добавления в excluded_keys: {params_to_add_to_excluded_keys}\n')

                prev_frame = traceback.extract_stack()[-2]
                self.add_excluded_keys(prev_frame, params_to_add_to_excluded_keys)
            else:
                raise exc


class BehaveTestCase(CheckJsonWithExclusionMixin, TestCase):

    live_server_url = '127.0.0.0'
    runner = None


class BehaveTransactionTestCase(CheckJsonWithExclusionMixin, TransactionTestCase):

    live_server_url = '127.0.0.0'

    def _fixture_setup(self):
        pass

    def _fixture_teardown(self):
        pass



from behave_bo.behave_django_bo.plugins.base import (
    Plugin,
)
from behave_bo.loggers import (
    tests_logger,
)


class OverrideSettingsPlugin(Plugin):
    """Указание значений параметров django settings для последующего переопределения при выполнении автотестов.

    $ behave --all --plugins override_settings --plugin_vars override_settings_PRODUCTION_REQUEST_LOG_SERVER=False
    """

    def __init__(self, *args, **kwargs):
        self.override_settings_dict = {}

        for key, value in kwargs.items():
            if value.isdigit():
                value = int(value)
            elif value.upper() in ('TRUE', "FALSE"):
                value = value.upper() == 'TRUE'

            self.override_settings_dict[key] = value

        tests_logger.info(f'Будут переопределены django settings: {self.override_settings_dict}')

    def before_all(self, context):
        context._runner.override_settings_dict = self.override_settings_dict


plugin_class = OverrideSettingsPlugin



"""
Provides a formatter that provides an overview of available step definitions
(step implementations).
"""

import inspect
from ast import (
    literal_eval,
)
from inspect import (
    getsourcelines,
)
from itertools import (
    groupby,
)
from operator import (
    attrgetter,
)
from typing import (
    List,
    Tuple,
)

from behave_bo import (
    i18n,
)
from behave_bo.formatter.base import (
    Formatter,
)
from behave_bo.matchers import (
    ParseMatcher,
)
from behave_bo.step_registry import (
    StepRegistry,
    registry,
)
from behave_bo.textutil import (
    compute_words_maxsize,
    indent,
    make_indentation,
    text as _text,
)
from behave_bo.userdata import (
    UserDataNamespace,
)


# -----------------------------------------------------------------------------
# CLASS: AbstractStepsFormatter
# -----------------------------------------------------------------------------
class AbstractStepsFormatter(Formatter):
    """
    Provides a formatter base class that provides the common functionality
    for formatter classes that operate on step definitions (implementations).

    .. note::
        Supports behave dry-run mode.
    """
    step_types = ("given", "when", "then", "step_template", "step",)

    def __init__(self, stream_opener, config):
        super().__init__(stream_opener, config)
        self.step_registry = None
        self.current_feature = None
        self.shows_location = config.show_source

    def reset(self):
        self.step_registry = None
        self.current_feature = None

    def discover_step_definitions(self):
        if self.step_registry is None:
            self.step_registry = StepRegistry()

        for step_type in registry.steps:
            if step_type == 'step':
                step_definitions = registry.steps[step_type]
                for _, step_definition in step_definitions.items():
                    step_definition.step_type = step_type
            else:
                step_definitions = tuple(registry.steps[step_type])
                for step_definition in step_definitions:
                    step_definition.step_type = step_type

            self.step_registry.steps[step_type] = step_definitions

    # -- FORMATTER API:
    def feature(self, feature):
        self.current_feature = feature
        if not self.step_registry:
            # -- ONLY-ONCE:
            self.discover_step_definitions()

    def eof(self):
        """Called at end of a feature."""
        self.current_feature = None

    def close(self):
        """Called at end of test run."""
        if not self.step_registry:
            self.discover_step_definitions()

        if self.step_registry:
            # -- ENSURE: Output stream is open.
            self.stream = self.open()
            self.report()

        # -- FINALLY:
        self.close_stream()

    # -- REPORT SPECIFIC-API:
    def report(self):
        raise NotImplementedError()

    # pylint: disable=no-self-use
    def describe_step_definition(self, step_definition, step_type=None):
        if not step_type:
            step_type = step_definition.step_type
        assert step_type
        return f"@{step_type}('{step_definition.pattern}')"


# -----------------------------------------------------------------------------
# CLASS: StepsFormatter
# -----------------------------------------------------------------------------
class StepsFormatter(AbstractStepsFormatter):
    """
    Provides formatter class that provides an overview
    which step definitions are available.

    EXAMPLE:
        $ behave --dry-run -f steps features/
        GIVEN STEP DEFINITIONS[21]:
          Given a new working directory
          Given I use the current directory as working directory
          Given a file named "{filename}" with
          ...
          Given a step passes
          Given a step fails

        WHEN STEP DEFINITIONS[14]:
          When I run "{command}"
          ...
          When a step passes
          When a step fails

        THEN STEP DEFINITIONS[45]:
          Then the command should fail with returncode="{result:int}"
          Then it should pass with
          Then it should fail with
          Then the command output should contain "{text}"
          ...
          Then a step passes
          Then a step fails

        GENERIC STEP DEFINITIONS[13]:
          * I remove the directory "{directory}"
          * a file named "{filename}" exists
          * a file named "{filename}" does not exist
          ...
          * a step passes
          * a step fails

    .. note::
        Supports behave dry-run mode.
    """
    name = "steps"
    description = "Shows step definitions (step implementations)."
    shows_location = True
    min_location_column = 40

    # -- REPORT SPECIFIC-API:
    def report(self):
        self.report_steps_by_type()

    def report_steps_by_type(self):
        """Show an overview of the existing step implementations per step type.
        """
        # pylint: disable=too-many-branches
        assert set(self.step_types) == set(self.step_registry.steps.keys())
        language = self.config.lang or "en"
        language_keywords = i18n.languages[language]

        for step_type in self.step_types:
            steps = list(self.step_registry.steps[step_type])
            if step_type != "step" and step_type != "step_template":
                steps.extend(self.step_registry.steps["step_template"])
                steps.extend(self.step_registry.steps["step"].values())

            if not steps:
                continue

            # -- PREPARE REPORT: For a step-type.
            step_type_name = step_type.upper()
            if step_type == "step":
                step_keyword = "*"
                step_type_name = "GENERIC"
            else:
                # step_keyword = step_type.capitalize()
                keywords = language_keywords[step_type]
                if keywords[0] == "*":
                    assert len(keywords) > 1
                    step_keyword = keywords[1]
                else:
                    step_keyword = keywords[0]

            steps_text = [f"{step_keyword} {step.pattern}"
                          for step in steps]
            if self.shows_location:
                max_size = compute_words_maxsize(steps_text)
                if max_size < self.min_location_column:
                    max_size = self.min_location_column
                schema = "  %-" + _text(max_size) + "s  # %s\n"
            else:
                schema = "  %s\n"

            # -- REPORT:
            message = "%s STEP DEFINITIONS[%s]:\n"
            self.stream.write(message % (step_type_name, len(steps)))
            for step, step_text in zip(steps, steps_text):
                if self.shows_location:
                    self.stream.write(schema % (step_text, step.location))
                else:
                    self.stream.write(schema % step_text)
            self.stream.write("\n")


# -----------------------------------------------------------------------------
# CLASS: StepsDocFormatter
# -----------------------------------------------------------------------------
class StepsDocFormatter(AbstractStepsFormatter):
    """
    Provides formatter class that shows the documentation of all registered
    step definitions. The primary purpose is to provide help for a test writer.

    EXAMPLE:
        $ behave --dry-run -f steps.doc features/
        @given('a file named "{filename}" with')
          Function: step_a_file_named_filename_with()
          Location: behave4cmd0/command_steps.py:50
            Creates a textual file with the content provided as docstring.

        @when('I run "{command}"')
          Function: step_i_run_command()
          Location: behave4cmd0/command_steps.py:80
            Run a command as subprocess, collect its output and returncode.

        @step('a file named "{filename}" exists')
          Function: step_file_named_filename_exists()
          Location: behave4cmd0/command_steps.py:305
            Verifies that a file with this filename exists.

            .. code-block:: gherkin

                Given a file named "abc.txt" exists
                 When a file named "abc.txt" exists
        ...

    .. note::
        Supports behave dry-run mode.
    """
    name = "steps.doc"
    description = "Shows documentation for step definitions."
    shows_location = True
    shows_function_name = True
    ordered_by_location = True
    doc_prefix = make_indentation(4)

    # -- REPORT SPECIFIC-API:
    def report(self):
        self.report_step_definition_docs()
        self.stream.write("\n")

    def report_step_definition_docs(self):
        step_definitions = []
        for step_type in self.step_types:
            if step_type == 'step':
                step_registry_definitions = self.step_registry.steps[step_type].values()
            else:
                step_registry_definitions = self.step_registry.steps[step_type]

            for step_definition in step_registry_definitions:
                # step_definition.step_type = step_type
                assert step_definition.step_type is not None
                step_definitions.append(step_definition)

        if self.ordered_by_location:
            step_definitions = sorted(step_definitions,
                                      key=attrgetter("location"))

        for step_definition in step_definitions:
            self.write_step_definition(step_definition)

    def write_step_definition(self, step_definition):
        step_definition_text = self.describe_step_definition(step_definition)
        self.stream.write("%s\n" % step_definition_text)
        doc = inspect.getdoc(step_definition.func)
        func_name = step_definition.func.__name__
        if self.shows_function_name and func_name not in ("step", "impl"):
            self.stream.write("  Function: %s()\n" % func_name)
        if self.shows_location:
            self.stream.write("  Location: %s\n" % step_definition.location)
        if doc:
            doc = doc.strip()
            self.stream.write(indent(doc, self.doc_prefix))
            self.stream.write("\n")
        self.stream.write("\n")


# -----------------------------------------------------------------------------
# CLASS: StepsCatalogFormatter
# -----------------------------------------------------------------------------
class StepsCatalogFormatter(StepsDocFormatter):
    """
    Provides formatter class that shows the documentation of all registered
    step definitions. The primary purpose is to provide help for a test writer.

    In order to ease work for non-programmer testers, the technical details of
    the steps (i.e. function name, source location) are ommited and the
    steps are shown as they would apprear in a feature file (no noisy '@',
    or '(', etc.).

    Also, the output is sorted by step type (Given, When, Then)

    Generic step definitions are listed with all three step types.

    EXAMPLE:
        $ behave --dry-run -f steps.catalog features/
        Given a file named "{filename}" with
            Creates a textual file with the content provided as docstring.

        When I run "{command}"
            Run a command as subprocess, collect its output and returncode.

        Given a file named "{filename}" exists
         When a file named "{filename}" exists
         Then a file named "{filename}" exists
            Verifies that a file with this filename exists.

            .. code-block:: gherkin

                Given a file named "abc.txt" exists
                 When a file named "abc.txt" exists
        ...

    .. note::
        Supports behave dry-run mode.
    """
    name = "steps.catalog"
    description = "Shows non-technical documentation for step definitions."
    shows_location = False
    shows_function_name = False
    ordered_by_location = False
    doc_prefix = make_indentation(4)

    def describe_step_definition(self, step_definition, step_type=None):
        if not step_type:
            step_type = step_definition.step_type
        assert step_type
        desc = []
        if step_type == "step":
            for step_type1 in self.step_types[:-1]:
                text = "%5s %s" % (step_type1.title(), step_definition.pattern)
                desc.append(text)
        else:
            desc.append(f"{step_type.title()} {step_definition.pattern}")

        return '\n'.join(desc)


# -----------------------------------------------------------------------------
# CLASS: StepsUsageFormatter
# -----------------------------------------------------------------------------
class StepsUsageFormatter(AbstractStepsFormatter):
    """
    Provides formatter class that shows how step definitions are used by steps.

    EXAMPLE:
        $ behave --dry-run -f steps.usage features/
        ...

    .. note::
        Supports behave dry-run mode.
    """
    name = "steps.usage"
    description = "Shows how step definitions are used by steps."
    doc_prefix = make_indentation(4)
    min_location_column = 40

    def __init__(self, stream_opener, config):
        super().__init__(stream_opener, config)
        self.step_usage_database = {}
        self.undefined_steps = []

    def reset(self):
        super().reset()
        self.step_usage_database = {}
        self.undefined_steps = []

    # pylint: disable=invalid-name
    def get_step_type_for_step_definition(self, step_definition):
        step_type = step_definition.step_type
        if not step_type:
            # -- DETERMINE STEP-TYPE FROM STEP-REGISTRY:
            assert self.step_registry
            for step_type, values in self.step_registry.steps.items():
                if step_definition in values:
                    return step_type
            # -- OTHERWISE:
            step_type = "step"
        return step_type
    # pylint: enable=invalid-name

    def select_unused_step_definitions(self):
        step_definitions = set()
        for step_type, values in self.step_registry.steps.items():
            if step_type == 'step':
                step_definitions.update(values.values())
            else:
                step_definitions.update(values)
        used_step_definitions = set(self.step_usage_database.keys())
        unused_step_definitions = step_definitions - used_step_definitions
        return unused_step_definitions

    def update_usage_database(self, step_definition, step):
        matching_steps = self.step_usage_database.get(step_definition, None)
        if matching_steps is None:
            assert step_definition.step_type is not None
            matching_steps = self.step_usage_database[step_definition] = []
        # -- AVOID DUPLICATES: From Scenario Outlines
        if not steps_contain(matching_steps, step):
            matching_steps.append(step)

    def update_usage_database_for_step(self, step):
        step_definition = self.step_registry.find_step_definition(step)
        if step_definition:
            self.update_usage_database(step_definition, step)
        # elif step not in self.undefined_steps:
        elif not steps_contain(self.undefined_steps, step):
            # -- AVOID DUPLICATES: From Scenario Outlines
            self.undefined_steps.append(step)

    # pylint: disable=invalid-name
    def update_usage_database_for_feature(self, feature):
        # -- PROCESS BACKGROUND (if exists): Use Background steps only once.
        if feature.background:
            for step in feature.background.steps:
                self.update_usage_database_for_step(step)

        # -- PROCESS SCENARIOS: Without background steps.
        for scenario in feature.walk_scenarios():
            for step in scenario.steps:
                self.update_usage_database_for_step(step)
    # pylint: enable=invalid-name

    # -- FORMATTER API:
    def feature(self, feature):
        super().feature(feature)
        self.update_usage_database_for_feature(feature)

    # -- REPORT API:
    def report(self):
        self.report_used_step_definitions()
        self.report_unused_step_definitions()
        self.report_undefined_steps()
        self.stream.write("\n")

    # -- REPORT SPECIFIC-API:
    def report_used_step_definitions(self):
        # -- STEP: Used step definitions.
        # ORDERING: Sort step definitions by file location.
        get_location = lambda x: x[0].location
        step_definition_items = self.step_usage_database.items()
        step_definition_items = sorted(step_definition_items, key=get_location)

        for step_definition, steps in step_definition_items:
            stepdef_text = self.describe_step_definition(step_definition)
            steps_text = [f"  {step.keyword} {step.name}"
                          for step in steps]
            steps_text.append(stepdef_text)
            max_size = compute_words_maxsize(steps_text)
            if max_size < self.min_location_column:
                max_size = self.min_location_column

            schema = "%-" + _text(max_size) + "s  # %s\n"
            self.stream.write(schema % (stepdef_text, step_definition.location))
            schema = "%-" + _text(max_size) + "s  # %s\n"
            for step, step_text in zip(steps, steps_text):
                self.stream.write(schema % (step_text, step.location))
            self.stream.write("\n")

    def report_unused_step_definitions(self):
        unused_step_definitions = self.select_unused_step_definitions()
        if not unused_step_definitions:
            return

        # -- STEP: Prepare report for unused step definitions.
        # ORDERING: Sort step definitions by file location.
        get_location = lambda x: x.location
        step_definitions = sorted(unused_step_definitions, key=get_location)
        step_texts = [self.describe_step_definition(step_definition)
                      for step_definition in step_definitions]

        max_size = compute_words_maxsize(step_texts)
        if max_size < self.min_location_column-2:
            max_size = self.min_location_column-2

        # -- STEP: Write report.
        schema = "  %-" + _text(max_size) + "s  # %s\n"
        self.stream.write("UNUSED STEP DEFINITIONS[%d]:\n" % len(step_texts))
        for step_definition, step_text in zip(step_definitions, step_texts):
            self.stream.write(schema % (step_text, step_definition.location))

    def report_undefined_steps(self):
        if not self.undefined_steps:
            return

        # -- STEP: Undefined steps.
        undefined_steps = sorted(self.undefined_steps,
                                 key=attrgetter("location"))

        steps_text = [f"  {step.keyword} {step.name}"
                      for step in undefined_steps]
        max_size = compute_words_maxsize(steps_text)
        if max_size < self.min_location_column:
            max_size = self.min_location_column

        self.stream.write("\nUNDEFINED STEPS[%d]:\n" % len(steps_text))
        schema = "%-" + _text(max_size) + "s  # %s\n"
        for step, step_text in zip(undefined_steps, steps_text):
            self.stream.write(schema % (step_text, step.location))


# -----------------------------------------------------------------------------
# CLASS: UnusedStepsRemoveFormatter
# -----------------------------------------------------------------------------
class UnusedStepsRemoveFormatter(StepsUsageFormatter):
    """Форматер для формирования отчета только по неиспользуемым шагам, с возможностью их удаления.

    $ behave --all --dry-run -f steps.remove_unused -D formatter.steps.remove_unused.repo_name=web-bb-salary \
        -D formatter.steps.remove_unused.with_removing=false
    """
    name = "steps.remove_unused"
    description = "Shows unused step definitions."
    userdata_scope = "formatter.steps.remove_unused"

    def __init__(self, stream_opener, config):
        super().__init__(stream_opener, config)
        self.unused_step_definitions = None
        self.with_removing = False
        self.repo_name = ''
        self.setup_with_userdata(config.userdata)

    def setup_with_userdata(self, userdata):
        config = UserDataNamespace(self.userdata_scope, userdata)
        self.with_removing = config.getbool("with_removing", self.with_removing)
        self.repo_name = config.get("repo_name", self.repo_name)

    def report(self):
        self.unused_step_definitions = self.select_unused_step_definitions()

        if not self.unused_step_definitions:
            self.stream.write(f'Unused step definitions not found.\n')
        else:
            self.report_unused_step_definitions()

        self.stream.write("\n")

    def report_unused_step_definitions(self):
        unused_step_definitions = sorted(self.unused_step_definitions, key=lambda x: x.location)
        self.stream.write(f'UNUSED STEP DEFINITIONS[{len(unused_step_definitions)}]:\n')

        for filepath, step_definitions in groupby(unused_step_definitions, key=lambda x: x.location.filename):
            if self.repo_name not in filepath:
                continue

            step_definitions = list(step_definitions)
            steps = list(self.step_registry.steps['step'].values())
            steps.extend(list(self.step_registry.steps['step_template']))

            left_count = len([step for step in steps if step.location.filename == filepath])

            slices, parse_failed_definitions = self.prepare_slices_to_remove(step_definitions)
            rm_count = len(step_definitions) - len(parse_failed_definitions)

            if self.with_removing and slices:
                self.remove_step_definitions(filepath, slices)
                self.stream.write(f'{filepath}:\n')
                self.stream.write(f'{rm_count} from {left_count} step definitions REMOVED.\n')
            else:
                self.stream.write(f'{filepath}\n')

                if parse_failed_definitions:
                    step_definitions = set(step_definitions).symmetric_difference(set(parse_failed_definitions))
                    step_definitions = sorted(step_definitions, key=lambda x: x.location.line)

                self.stream.write(f'{rm_count} from {left_count} step definitions CAN BE REMOVED.\n')

                for definition in step_definitions:
                    self.stream.write(f'{definition.location}\n')

            if parse_failed_definitions:
                self.stream.write(f'{len(parse_failed_definitions)} step definitions FAILED TO PARSE:\n')
                for definition in parse_failed_definitions:
                    self.stream.write(f'{definition.pattern}\n')
                    self.stream.write(f'{definition.location}\n')

            self.stream.write('\n')

    def remove_step_definitions(self, filepath: str, slices: List[slice]) -> None:
        """Удаляет строки кода из py файла, содержащего определения шагов.

        Args:
            filepath: путь до файла.
            slices: срезы для удаления.

        """
        with open(filepath) as f:
            file_lines = f.readlines()

        for sl in sorted(slices, reverse=True, key=lambda s: s.start):
            file_lines[sl] = []

        code_separator = '\n\n\n'
        file_parts = ''.join(file_lines).split(code_separator)
        # Удаление пустых строк между блоками кода.
        file_parts = filter(bool, (part.strip() for part in file_parts))
        result_file_text = code_separator.join(file_parts) + '\n'

        with open(filepath, 'w') as f:
            f.write(result_file_text)

    def prepare_slices_to_remove(self, step_definitions: List[ParseMatcher]) -> Tuple[List[slice], List[ParseMatcher]]:
        """Формирует срезы для удаления из файла.

        Args:
            step_definitions: Список определений шагов.

        Returns:
            (Список срезов, список не найденных шагов).

        """
        slices = []
        not_found_definitions = []

        for func, step_definitions in groupby(step_definitions, key=lambda x: x.func):
            step_definitions = list(step_definitions)
            group_len = len(step_definitions)
            func_lines, first_line_num = getsourcelines(func)
            first_line_index = first_line_num - 1
            step_decorators_count = len([line for line in func_lines if line.startswith('@step(') or line.startswith('@step_template(')])

            if group_len == step_decorators_count:
                last_line_index = first_line_index + len(func_lines)
                removing_slice = slice(first_line_index, last_line_index)
                slices.append(removing_slice)
            else:
                definitions_pattern_map = self.make_definitions_pattern_map(func_lines, first_line_index)
                for step_definition in step_definitions:
                    pattern = step_definition.pattern
                    if pattern in definitions_pattern_map:
                        slices.append(slice(*definitions_pattern_map[pattern]))
                    else:
                        not_found_definitions.append(step_definition)

        return slices, not_found_definitions

    def make_definitions_pattern_map(self, func_lines: List[str], offset: int) -> dict:
        """Формирует словарь, где ключ - шаблон шага из декоратора step,
        значение - пара индексов (start, end) расположения строк декоратора в коде модуля.

        Args:
            func_lines: Список строк функции шага.
            offset: Смещение (индекс первой строки функции в коде модуля).

        Returns:
            Словарь соответствия шаблона расположению в коде.

        """
        indices = self.get_decorator_indices(func_lines)
        pattern_map = {}
        for start, end in indices:
            pattern = literal_eval(''.join(func_lines[start:end]).replace('@step_template', '').replace('@step', ''))
            pattern_map[pattern] = start + offset, end + offset

        return pattern_map

    def get_decorator_indices(self, lines: List) -> List[Tuple[int, int]]:
        """Формирует список индексов расположения строк декоратора в коде функции.

        Args:
            lines: Список строк функции шага.

        Returns:
            Список пар индексов.

        """
        start_indices = [i for i, string in enumerate(lines) if string.startswith('@step(') or string.startswith('@step_template(')]
        end_indices = []
        for i in start_indices:
            for index, line in enumerate(lines[i + 1:], i + 1):
                if line.startswith('@step(') or line.startswith('@step_template(') or line.startswith('#') or line.startswith('def '):
                    end_indices.append(index)
                    break

        return list(zip(start_indices, end_indices))


# -----------------------------------------------------------------------------
# UTILITY FUNCTIONS:
# -----------------------------------------------------------------------------
def steps_contain(steps, step):
    for other_step in steps:
        if step == other_step and step.location == other_step.location:
            # -- NOTE: Step comparison does not take location into account.
            return True
    # -- OTHERWISE: Not contained yet (or step in other location).
    return False



import os
import zipfile
from difflib import (
    unified_diff,
)
from tempfile import (
    TemporaryDirectory,
)

from django.conf import (
    settings,
)

from behave_bo.model import (
    Table,
)


def get_context_extractor_class():
    # TODO BOBUH-23698 Вынести настройку CONTEXT_EXTRACTOR из web_bb_behave
    context_extractor_cls = (
        settings.WEB_BB_BEHAVE__CONTEXT_EXTRACTOR or
        'behave_bo.behave_django_bo.extractors.ContextExtractor'
    )
    context_extractor_cls_path = context_extractor_cls.split('.')
    # Allow for relative paths
    if len(context_extractor_cls_path) > 1:
        extractor_module_name = '.'.join(context_extractor_cls_path[:-1])
    else:
        extractor_module_name = '.'
    extractor_module = __import__(extractor_module_name, {}, {}, context_extractor_cls_path[-1])
    return getattr(extractor_module, context_extractor_cls_path[-1])


def filter_dict_keys(
    obj,
    excluded_keys,
):
    """Исключает ключи из словаря и его вложенных словарей.

    Args:
        obj: Словарь/список/любое значение.
        excluded_keys: Список строк ключей для исключения.
    Returns:
        Словарь/список/любое значение без исключённых ключей.
    """
    if isinstance(obj, dict):
        obj = {
            key: filter_dict_keys(
                value,
                excluded_keys,
            )
            for key, value in obj.items()
            if key not in excluded_keys
        }
    elif isinstance(obj, list):
        obj = [
            filter_dict_keys(
                item,
                excluded_keys,
            )
            for item in obj
        ]

    return obj


def sort_lists_in_dict(obj, sort_key=None):
    """Сортирует все списки в словаре.

    Args:
        data: Словарь/список/любое значение.
        sort_key: Ключ для сортировки списка.

    Returns:
        Словарь/список/любое значение с отсортированными списками.

    """
    if isinstance(obj, dict):
        for key, value in obj.items():
            obj[key] = sort_lists_in_dict(value, sort_key)

    elif isinstance(obj, list):
        for item in obj:
            sort_lists_in_dict(item, sort_key)

        obj.sort(key=sort_key)

    return obj


class FilesCompare:
    """Предоставляет функционал сравнения файлов.

    Attributes:
        message: str. Сообщение с подробным результатом сравнения.
        compare_file_names: bool. Сравнивать наименования файлов.
            во временной директории

    """

    TABLE_MAP = {
        'Без проверки имен файлов': 'without_compare_file_names',
    }

    def __init__(self, **kwargs):
        self.message = ''
        self.compare_file_names = kwargs.get('compare_file_names', True)
        self._in_temporary_dir = False

        if kwargs.get('table'):
            params = self._processing_table(kwargs['table'])
            self._processing_params(params)
        else:
            self._processing_params(kwargs)

        self._compare_func_type_map = {
            'CSV': self.compare_csv_files,
            'TXT': self.compare_text_files,
            'HTML': self.compare_html_files,
        }

    def _processing_table(self, table: Table) -> dict:
        """Преобразует behave таблицу в словарь, при этом наименования столбцов
        заменяет в соответствии с таблицей TABLE_MAP.

        Args:
            table: объект Table полученный из *feature файла*.

        Returns:
            Словарь параметров.
        """

        return {self.TABLE_MAP[k]: v for k, v in table.rows[0].items() if k in self.TABLE_MAP.keys()}

    def _processing_params(self, params: dict) -> None:
        """Выполняет обработку параметров сравнения.

        Args:
            params: Словарь параметров.

        Returns:
            None.
        """
        if 'without_compare_file_names' in params:
            self.compare_file_names = False

    def compare_csv_files(self, left: str, right: str):
        """Сравнивает csv файлы.

        Args:
            left: Абсолютный путь до файла.
            right: Абсолютный путь до файла.

        Returns:
            Результат сравнения.
        """

        return self.compare_text_files(left, right, encoding='utf8')

    def compare_text_files(self, left: str, right: str, encoding: str = 'cp1251'):
        """Сравнивает текстовые файлы.

        Args:
            left: Абсолютный путь до файла.
            right: Абсолютный путь до файла.
            encoding: Кодировка файлов.

        Returns:
            Результат сравнения.
        """
        with open(left, encoding=encoding) as left_file:
            with open(right, encoding=encoding) as right_file:
                left_lines = left_file.readlines()
                right_lines = right_file.readlines()

                if len(left_lines) != len(right_lines):
                    result = False
                    self.message += 'Файлы содержат разное количество строк.\n'
                else:
                    diff = list(unified_diff(left_lines, right_lines, 'Left', 'Right'))
                    result = not diff
                    self.message += f'{"".join(diff)}'
        if result:
            self.message += 'OK\n'

        return result

    def compare_html_files(self, left: str, right: str):
        """Сравнивает html файлы.

        Args:
            left: Абсолютный путь до файла.
            right: Абсолютный путь до файла.

        Returns:
            Результат сравнения.
        """

        return self.compare_text_files(left, right, encoding='utf8')

    def compare_by_type(self, left: str, right: str, file_type: str = None) -> bool:
        """Вызывает функцию сравнения в соответствии с переданным типом.
        Если передан неизвестный тип, будет вызвано сравнение текстовых файлов.

        Args:
            left: Абсолютный путь до файла.
            right: Абсолютный путь до файла.
            file_type: Тип файла.

        Returns:
            Результат выполнения функции.
        """
        if not file_type:
            file_type = self.get_type(left)
        try:
            compare_func = self._compare_func_type_map[file_type]
        except KeyError:
            result = self.compare_text_files(left, right)
        else:
            result = compare_func(left, right)

        return result

    def compare_zip_archives(self, left: str, right: str) -> bool:
        """Сравнивает zip архивы.

        Args:
            left: Абсолютный путь до файла.
            right: Абсолютный путь до файла.

        Returns:
            Результат сравнения.
        """
        with TemporaryDirectory() as left_tmp_dir:
            with TemporaryDirectory() as right_tmp_dir:
                self._in_temporary_dir = True
                left_zip_archive = zipfile.ZipFile(left)
                left_zip_archive.extractall(left_tmp_dir)
                right_zip_archive = zipfile.ZipFile(right)
                right_zip_archive.extractall(right_tmp_dir)
                result = self.compare_dirs(left_tmp_dir, right_tmp_dir)

        return result

    def compare_dirs(self, left_dir: str, right_dir: str) -> bool:
        """Сравнивает директории по файлам без учета поддиректорий.

        Args:
            left_dir: Абсолютный путь.
            right_dir: Абсолютный путь.

        Returns:
            Результат сравнения.
        """
        left_files = sorted(os.listdir(left_dir))
        right_files = sorted(os.listdir(right_dir))

        if len(left_files) != len(right_files):
            result = False
            self.message = (
                f'Директории содержат разное количество файлов:\n'
                f'Left:  {left_files}\nRight: {right_files}\n'
            )
        elif self.compare_file_names and left_files != right_files:
            result = False
            self.message = (
                f'Директории отличаются содержанием:\n'
                f'Left:  {left_files}\nRight: {right_files}\n'
            )
        else:
            result = True
            for left, right in zip(left_files, right_files):
                self.message += f'Сравнение файлов {left} и {right}.\n'
                temp_result = self.compare_by_type(
                    os.path.join(left_dir, left),
                    os.path.join(right_dir, right),
                    self.get_type(left)
                )
                result = result and temp_result

        return result

    def get_type(self, path_name: str) -> str:
        """Получает тип файла по расширению файла.
        Если не указано расширение, возвращает TXT.

        Args:
            path_name: Путь или наименование файла

        Returns:
            Строковый тип файла.
        """
        _, extension = os.path.splitext(path_name)
        if extension:
            file_type = extension[1:].upper()
        else:
            file_type = 'TXT'

        return file_type


context_extractor_class = get_context_extractor_class()
context_extractor = context_extractor_class()



from datetime import (
    datetime,
)

try:
    from celery.signals import (
        task_failure,
    )
except ImportError:
    task_failure = None

from django.db import (
    transaction,
)
from django.dispatch import (
    receiver,
)
from factory.django import (
    mute_signals,
)

from behave_bo.behave_django_bo.cache import (
    django_cache_dump,
    django_cache_restore,
    setup_cache_server_db,
)
from behave_bo.behave_django_bo.client import (
    BehaveBOClient,
)
from behave_bo.behave_django_bo.enums import (
    SaveDbParameterEnum,
)
from behave_bo.behave_django_bo.plugins.base import use_plugins
from behave_bo.behave_django_bo.requests import (
    AppRequest,
)
from behave_bo.behave_django_bo.signals import (
    after_scenario_signal,
    after_step_plugins,
)
from behave_bo.behave_django_bo.utils import (
    ModelRecordAnalyzer,
)
from behave_bo.consts import (
    scenario_tag_cookie_key,
)
from behave_bo.loggers import (
    tests_logger,
)


failure_task_exceptions = []

if task_failure:
    @receiver(task_failure)
    def handle_task_failure(exception, **kwargs):
        """
        Обработка сигнала ошибочного завершения фоновой, для последующей
        обработки в process_after_step.
        """
        failure_task_exceptions.append(exception)


class Environment:

    def failure_task_possible_exception(self, exc, context):
        return f'ignore{exc.__class__.__name__}' in context.scenario.tags

    def process_after_step(self, context):
        """
        Некоторые таски celery запускаются через событие on_commit,
        которое не происходит когда нужно, потому что каждый тест
        оборачивается в atomic. Прогоним вручную
        """
        conn = transaction.get_connection()

        if conn.in_atomic_block:
            for _, _func in conn.run_on_commit:
                if signals_to_mute := context.signals_to_mute_in_tasks:
                    with mute_signals(*signals_to_mute):
                        _func()
                else:
                    _func()

                if failure_task_exceptions:
                    raise failure_task_exceptions.pop()

            conn.run_on_commit = []

        if failure_task_exceptions:
            raw_exceptions = []
            for exc in failure_task_exceptions:
                if self.failure_task_possible_exception(exc, context):
                    tests_logger.info(f'{exc.__class__}: {exc}')
                else:
                    raw_exceptions.append(exc)

            failure_task_exceptions.clear()

            if raw_exceptions:
                raw_exceptions_str = '\n'.join(f'{e.__class__.__name__}: {e}' for e in raw_exceptions)
                raise Exception(
                    f'Количество необработанных ошибок в фоновых задачах: {len(raw_exceptions)}\n{raw_exceptions_str}'
                )

    @use_plugins
    def before_all(self, context):
        context.session = BehaveBOClient()
        context.AppRequest = AppRequest

        if context.config.parallel:
            setup_cache_server_db(context.config.group_number)

    @use_plugins
    def before_feature(self, context, feature):
        pass

    @use_plugins
    def before_scenario(self, context, scenario):
        if f'{scenario.main_tag}' in context.config.analyze_model_records:
            context.model_record_analyzer = ModelRecordAnalyzer(context)

        context.today_date = datetime.today()

        # Сохраняем DjangoCache
        context.cache_dump = django_cache_dump()

        # Инициализируем объект теста
        context.test = context.config.test_case_type()

        # Отключаем лимит вывода информации от assert-методов сравнения,
        # в случае несоответствия сравниваемых данных.
        context.test.maxDiff = None

        # инициализируем тесткейс
        setup_result = context.test.setUpClass()
        if setup_result is False:
            scenario.skip("Can't setup class")

        test_tag = scenario.main_tag
        context.test_tag = test_tag

        if context.config.save_db == SaveDbParameterEnum.SEPARATE:
            if scenario._row:
                test_tag += scenario._row.id

            context.config.django_test_runner.setup_databases(
                db_postfix=test_tag,
            )

        self.set_session_cookies_scenario_tags(context, scenario)

    @use_plugins
    def before_step(self, context, step):
        setattr(context, 'current_step', step)
        context.signals_to_mute_in_tasks = None

    @use_plugins
    def after_step(self, context, step):
        after_step_plugins.send(sender=None, context=context, step=step)
        
        self.process_after_step(context)

        if hasattr(context, 'current_step'):
            delattr(context, 'current_step')

    @use_plugins
    def after_scenario(self, context, scenario):
        context.test.tearDownClass()

        if context.config.save_db == None:
            context.config.django_test_runner.reset_sequences()

        del context.test

        # Восстанавливаем DjangoCache
        django_cache_restore(context.cache_dump)
        after_scenario_signal.send(sender=None, context=context, scenario=scenario)

        if f'{scenario.main_tag}' in context.config.analyze_model_records:
            context.model_record_analyzer.write()

        if context.config.log_non_200_requests:
            context.session.print_non_200_requests()

    @use_plugins
    def after_feature(self, context, feature):
        """Объявление функции нужно для выполнения кода в соответствующем хуке плагинов."""
        pass

    @use_plugins
    def after_all(self, context):
        context.session = None
        context.config.django_test_runner.teardown_database()

    def set_session_cookies_scenario_tags(self, context, scenario):
        """
        Прокидывает в cookies сессии информацию о тегах сценария, -- указывает основной тег сценария.
        Для случая Структуры сценария дополняет тег сценария номером примера.

        Args:
            context: объект контекста тестирования behave_bo.runner.Context
            scenario: объект сценария behave_bo.model.Scenario
        """
        if scenario._row:
            cookie_test_tag = f'{scenario.main_tag}-{scenario._row.id}'
        else:
            cookie_test_tag = scenario.main_tag

        context.session.cookies[scenario_tag_cookie_key] = cookie_test_tag


environment_object = Environment()



import datetime
import os
import traceback
import uuid
from collections import (
    defaultdict,
)
from typing import (
    List,
    Tuple,
)

from django.conf import (
    settings,
)
from django.db import (
    connections,
)
from django.db.models.query_utils import (
    DeferredAttribute,
)
from django.db.utils import (
    load_backend,
)
from django.test.utils import (
    override_settings,
)

from behave_bo.behave_django_bo.base_runner import (
    BaseBehaveTestRunner,
)
from behave_bo.behave_django_bo.consts import (
    override_settings_prefix,
    replace_current_date_prefix,
)
from behave_bo.behave_django_bo.loggers import (
    FixChecksLogger,
)
from behave_bo.behave_django_bo.plugins.base import (
    RunnerPluginsMixin,
)
from behave_bo.behave_django_bo.testcase import (
    BehaveTestCase,
    BehaveTransactionTestCase,
)
from behave_bo.behave_django_bo.utils import (
    override_current_date,
    override_dirs,
)
from behave_bo.behave_django_bo.wrapper_with_queries_log import (
    changed_tables,
)
from behave_bo.configuration import (
    Configuration,
)
from behave_bo.loggers import (
    tests_logger,
)
from behave_bo.model import (
    Scenario,
)


behave_run_scenario = Scenario.run


class PatchedScenario(Scenario):

    @property
    def override_settings_tag(self):
        override_settings_tag = None

        for tag in self.tags:
            if tag.startswith(override_settings_prefix):
                override_settings_tag = tag
                break

        return override_settings_tag

    def make_override_settings_dict(self, runner) -> dict:
        """Формирует словарь настроек для переопределения в override_settings.

        Настройки, задаваемые в runner.override_settings_dict, будут переопределены
        или дополнены для конкретного сценария.

        Args:
            runner: объект класса Runner, исполнителя behave-автотестов.

        Returns:
            Словарь вида {Наименование настройки: значение}.

        """
        override_settings_dict = getattr(runner, 'override_settings_dict', {})

        if self.override_settings_tag:
            settings_string = self.override_settings_tag.removeprefix(override_settings_prefix)
            for key, value in map(lambda s: s.split('=', 1), settings_string.split(',')):
                if value.isdigit():
                    value = int(value)
                elif value.lower() in ('true', 'false'):
                    value = value.lower() == 'true'
                elif value.lower() == 'none':
                    value = None

                override_settings_dict[key] = value

            tests_logger.info(
                f'Для сценария {self.main_tag} будут переопределены django settings: {override_settings_dict}'
            )

        return override_settings_dict

    @property
    def replace_current_date_tag(self):
        """Определение тега замены текущей даты у сценария."""
        replace_current_date_tag = None

        for tag in self.tags:
            if tag.startswith(replace_current_date_prefix):
                replace_current_date_tag = tag
                break

        return replace_current_date_tag

    def get_scenario_current_date(self) -> datetime.date:
        """Получение даты из тега замены текущей даты сценария.

        Returns:
            Объект даты
        """
        replace_current_date = None

        if self.replace_current_date_tag:
            replace_current_date = datetime.datetime.strptime(
                self.replace_current_date_tag.removeprefix(replace_current_date_prefix).split('=')[1],
                '%d.%m.%Y',
            ).date()

            tests_logger.info(
                f'Для сценария {self.main_tag} будут переопределена текущая дата: {replace_current_date:%d.%m.%Y}'
            )

        return replace_current_date

    def run_scenario(self, runner):
        """Пропатченный метод запуска behave-сценария.

        Сразу проверяется необходимость выполнения сценария, и если не нужно выполнять, помечается пропущенным.
        Если нужно выполнять, на время выполнения конкретного сценария создаёт уникальную временную директорию.

        Args:
            self: объект класса Scenario
            runner: объект класса Runner, исполнителя behave-автотестов.
        """
        failed = False
        should_run_scenario = self.should_run(runner.config)

        if should_run_scenario:
            for number, step in enumerate(self.steps, 1):
                step.number = number
            settings_dict = self.make_override_settings_dict(runner)
            scenario_current_date = self.get_scenario_current_date()

            with override_dirs(self), override_settings(**settings_dict), override_current_date(scenario_current_date):
                failed = behave_run_scenario(self, runner)
        else:
            self.mark_skipped()

        return failed


Scenario.base_skip_tags = list(Scenario.base_skip_tags) + [override_settings_prefix]
Scenario.override_settings_tag = PatchedScenario.override_settings_tag
Scenario.replace_current_date_tag = PatchedScenario.replace_current_date_tag
Scenario.make_override_settings_dict = PatchedScenario.make_override_settings_dict
Scenario.get_scenario_current_date = PatchedScenario.get_scenario_current_date
Scenario.run = PatchedScenario.run_scenario


class BehaveTestRunner(BaseBehaveTestRunner, RunnerPluginsMixin):

    reset_sequences_sql = {}
    default_db_name = None
    test_db_name = None
    behave_config: Configuration = None
    max_identifier_length: int = None

    def get_tempdir_path(self):
        return settings.TEMPDIR

    def get_template_data_types(self):
        return dict()

    def get_environment_filepath(self):
        return os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            'environment.py'
        )

    def define_testcase_class(self, config):
        if config.save_changes_to_db:
            testcase_class = BehaveTransactionTestCase
        else:
            testcase_class = BehaveTestCase

        return testcase_class

    def run_behave_tests(self, behave_args, options):
        with FixChecksLogger(options.get('fix_checks', False) or options.get('fix_excluded_keys', False)):
            run_result = super().run_behave_tests(behave_args, options)

        self._remove_readonly_requests_files()

        return run_result

    def _add_extra_parameters_to_behave_config(self, options):
        """Добавляет дополнительные параметры в конфиг в соответствии с options.

        Args:
            options: Аргументы команды запуска.

        """
        config = super()._add_extra_parameters_to_behave_config(options)

        config.bgjob_wait_time = options.get('bgjob_wait_time')
        config.bgjob_wait_iteration_freeze = options.get('bgjob_wait_iteration_freeze')
        config.remove_readonly_requests = options.get('remove_readonly_requests')
        config.remove_readonly_requests_data = {
            'not_used_request_files': set(),
            'not_used_steps_lines': defaultdict(list),
            'used_request_files': set(),
            'used_steps_funcs': defaultdict(set),
        }
        config.skip_readonly_requests = options.get('skip_readonly_requests')
        config.skip_readonly_requests_url_endings = options.get('skip_readonly_requests_url_endings')
        config.skip_production_request_checks = options.get('skip_production_request_checks')
        config.fix_checks = options.get('fix_checks')
        config.fix_excluded_keys = options.get('fix_excluded_keys')
        config.analyze_model_records = options.get('analyze_model_records', '').split(',')
        config.log_non_200_requests = options.get('log_non_200_requests')

        config.test_case_type = self.define_testcase_class(config)
        config.test_case_type.runner = self

        if config.save_changes_to_db and not options.get('tags'):
            raise Exception(
                'Сохранение БД возможно только при запуске индивидуальных тестов!'
            )
        if config.save_changes_to_db and config.parallel:
            raise Exception(
                'Нельзя одновременно использовать параметры --parallel и --save-db'
            )
        if config.fix_excluded_keys and config.fix_checks:
            raise Exception(
                'Нельзя одновременно использовать параметры --fix-checks и --fix-excluded-keys'
            )

        return config

    def setup_databases(self, **kwargs):
        """
        Создание БД перед тестированием - средствами SQL и частично Django
        """
        connections['default'].close()
        db_postfix = kwargs.get('db_postfix') or uuid.uuid4().hex
        self.test_db_name = f'_{db_postfix}'

        self.default_db_name = settings.DATABASES['default']['NAME']
        temp_settings = connections['default'].settings_dict.copy()

        backend = load_backend(temp_settings['ENGINE'])
        temp_settings['NAME'] = 'postgres'
        temp_connection = backend.DatabaseWrapper(
            temp_settings,
            alias='temp_connection',
        )

        cursor = temp_connection.cursor()

        try:
            cursor.connection.set_isolation_level(0)
            cursor.execute(f'DROP DATABASE IF EXISTS "{self.test_db_name}"')
        except Exception as e:
            tests_logger.critical(f"Database deletion error: {e}\n")

        try:
            cursor.execute(f'CREATE DATABASE "{self.test_db_name}" '
                           f'TEMPLATE {self.default_db_name}')
        except Exception as e:
            tests_logger.critical(f'Database creation error: {e}\n')
            exit(-1)

        tests_logger.info(f'Database "{self.test_db_name}" created')
        temp_settings['NAME'] = self.test_db_name
        connections['default'] = backend.DatabaseWrapper(
            temp_settings,
            alias='default',
        )
        temp_connection.close()
        connections['default'].connect()

    def teardown_database(self):
        """
        Снос копии БД созданной для запуска автотестов
        """
        if not self.keepdb and self.test_db_name:
            connections['default'].close()
            db_settings = connections['default'].settings_dict.copy()
            backend = load_backend(db_settings['ENGINE'])
            db_settings['NAME'] = self.default_db_name
            default_connection = connections['default'] = backend.DatabaseWrapper(
                db_settings,
                alias='default',
            )
            default_connection.connect()

            cursor = default_connection.cursor()
            try:
                cursor.execute(f'DROP DATABASE IF EXISTS "{self.test_db_name}"')
                tests_logger.info(f'Database "{self.test_db_name}" destroyed!')
            except Exception as e:
                tests_logger.critical(f"Database deletion error: {e}\n")
            else:
                self.test_db_name = None

    def get_sequences(self, connection) -> List[Tuple[str, str]]:
        """Получает список последовательностей таблиц бд.

        Args:
            connection: Конект к бд.

        Returns:
            Список наименований таблиц и их последовательностей.
        """
        with connection.cursor() as cursor:
            cursor.execute(
                """SELECT tbl.relname, s.relname as sequence_name 
                FROM pg_class s
                    JOIN pg_depend d ON d.refobjid = s.oid AND d.refclassid = 'pg_class'::regclass
                    JOIN pg_attrdef ad ON ad.oid = d.objid AND d.classid = 'pg_attrdef'::regclass
                    JOIN pg_class tbl ON tbl.oid = ad.adrelid
                    JOIN pg_namespace n ON n.oid = tbl.relnamespace
                WHERE s.relkind = 'S'
                    AND d.deptype in ('a', 'n')
                    AND n.nspname = 'public';"""
            )
            return cursor.fetchall()

    def define_max_identifier_length(self, connection):
        """Определяет и сохраняет максимальную длину идентификатора бд max_identifier_length.

        Args:
            connection: Конект к бд.
        """
        with connection.cursor() as cursor:
            cursor.execute('SHOW max_identifier_length;')
            self.max_identifier_length = int(cursor.fetchone()[0])

    def collect_reset_sequences_sql(self) -> None:
        """Собирает данные по последовательностям бд.

        Заносит в self.reset_sequences_sql наименование таблицы и sql-код для сброса последовательности.
        """
        connection = connections['default']
        self.define_max_identifier_length(connection)
        for table, sequence in self.get_sequences(connection):
            with connection.cursor() as cursor:
                cursor.execute(
                    f'SELECT last_value, is_called from "{sequence}";'
                )
                last_value, is_called = cursor.fetchone()
                setval_sql = (
                    f"SELECT setval('\"{sequence}\"', {last_value}, {is_called});"
                )
                self.reset_sequences_sql[table] = setval_sql

    def reset_sequences(self) -> None:
        """Сбрасывает последовательности таблиц."""
        connection = connections['default']
        with connection.cursor() as cursor:
            for table in changed_tables.copy():
                pg_table_name = table[:self.max_identifier_length]
                if pg_table_name in self.reset_sequences_sql:
                    cursor.execute(self.reset_sequences_sql[pg_table_name])
                changed_tables.remove(table)

    def collect_requests_data(self, is_used) -> None:
        """
        Найдем из стека вызовов данные о используемых/неиспользуемых файлах отправки запросов.
        Берём последние 4 фрейма т.к. стандартно вызов отправки запроса происходит из функции шага,
        далее в функции запроса, далее в методе execute, и далее в текущем методе.

        Args:
            is_used: Признак использования файла запроса
        """
        last_frames_index = -4
        request_file_prefix = 'request_'
        steps_file_prefix = 'steps_'
        data = self.behave_config.remove_readonly_requests_data

        for frame in traceback.extract_stack()[last_frames_index:]:
            filename = os.path.basename(frame.filename)

            if is_used:
                if filename.startswith(steps_file_prefix):
                    data['used_steps_funcs'][frame.filename].add(
                        frame.line.split('(')[0],  # название функции отправки запроса
                    )
                elif filename.startswith(request_file_prefix):
                    data['used_request_files'].add(frame.filename)
            else:
                if filename.startswith(steps_file_prefix):
                    data['not_used_steps_lines'][frame.filename].append((
                        frame.lineno,  # номер линии в файле, где вызывается функция
                        frame.line.split('(')[0],  # название функции отправки запроса
                    ))
                elif filename.startswith(request_file_prefix):
                    data['not_used_request_files'].add(frame.filename)

    def _remove_readonly_requests_files(self):
        """
        Удаляет файлы readonly-запросов шагов предыстории из автотестов на основе данных remove_readonly_requests_data.
        """
        data = self.behave_config.remove_readonly_requests_data
        remove_key = 'remove'

        for step_file, lines_list in data['not_used_steps_lines'].items():
            tests_logger.info(f'Удаляем вызов функции отправки readonly-запроса из файла реализации шагов {step_file}')
            used_steps_funcs = data['used_steps_funcs'][step_file]

            with open(step_file) as f:
                lines = f.readlines()

            remove_indexes, remove_func_names = zip(*lines_list)

            for idx, line in enumerate(lines):
                file_line_num = idx + 1

                if file_line_num in remove_indexes:
                    # Удаление из функции реализации шага
                    lines[idx] = remove_key
                elif any(f'{remove_func_name},' in line and remove_func_name not in used_steps_funcs
                         for remove_func_name in remove_func_names):
                    # Удаление из секции импортов
                    for remove_idx in (
                        idx - 1,
                        idx,
                        idx + 1,
                    ):
                        lines[remove_idx] = remove_key

            new_lines = []
            step_def = False

            for line in lines:
                if line != remove_key:
                    if step_def:
                        if not line.strip():
                            new_lines.append('    pass\n')
                        step_def = False

                    new_lines.append(line)

                if 'def ' in line:
                    step_def = True

            with open(step_file, 'w') as f:
                f.writelines(new_lines)

        for request_file in (data['not_used_request_files'] - data['used_request_files']):
            tests_logger.info(f'Удаляем файл readonly-запроса {request_file}')
            try:
                os.remove(request_file)
            except Exception as e:
                tests_logger.info(f'Ошибка при удалении файла: {e}')


def _DeferredAttributeWithCheck__get__(self, instance, cls=None):
    """
    Добавление проверки/логирования в метод отложенного получения значения атрибута.

    Args:
        self: экземпляр класса DeferredAttribute
        instance: экземпляр класса модели для которого выполняется получение значения атрибута.
        cls: класс модели

    Returns:
        Значение атрибута.

    Raises:
        Исключение в случае если включена настройка CHECK_DEFERRED_ATTR_GET
         и значение атрибута не найдено в кэше, а значит его планируется получить из БД выполнив дополнительный запрос.
    """
    if instance is None:
        return self
    data = instance.__dict__
    field_name = self.field.attname

    if field_name in data:
        return data[field_name]

    attr = f"{instance.__class__.__name__}.{field_name}"
    message = f"Lazy fetching of {attr} may cause 1+N issue"

    if settings.WEB_BB_BEHAVE__CHECK_DEFERRED_ATTR_GET:
        raise AssertionError(message)
    elif settings.WEB_BB_BEHAVE__LOG_DEFERRED_ATTR_GET:
        tests_logger.warning(message)

    return DeferredAttribute__get(self, instance, cls)


DeferredAttribute__get, DeferredAttribute.__get__ = DeferredAttribute.__get__, _DeferredAttributeWithCheck__get__



import re

from django.db.backends import (
    utils,
)
from django.db.backends.postgresql import (
    base,
)


changed_tables = set()


class CursorWrapper(utils.CursorWrapper):
    """
    Враппер курсора БД для получения информации об операция вставки в таблицы
    """
    sql_insert_operator = "INSERT INTO"
    re_pattern = re.compile(f'{sql_insert_operator} "?(.*?)"? .+', re.U | re.I)

    def execute(self, sql, params=None):
        if self.sql_insert_operator in sql.upper():
            result = self.re_pattern.search(sql)
            if result:
                changed_tables.add(result[1])


class DatabaseWrapper(base.DatabaseWrapper):

    def cursor(self):
        self.validate_thread_sharing()
        cursor = CursorWrapper(self._cursor(), self)

        return cursor



import argparse
import ast
import os
import re
import uuid
from collections import (
    defaultdict,
)
from pathlib import (
    Path,
)


class Storage:
    def __init__(self, processor_ins):
        self._processor_ins = processor_ins
        self.storage = defaultdict(dict)

    @property
    def project(self):
        return self._processor_ins.project_dir

    @property
    def records_in_project(self):
        return self.storage[self.project]

    def get(self, key):
        return self.records_in_project[key]


class FactoriesReuseProcessor:

    def __init__(self, base_path) -> None:
        self.base_path = base_path
        self.project_dir = ''

        self.factories_records = Storage(self)
        self.factories_imports = Storage(self)
        self.all_func_names = set()
        self.records_file_paths_for_factories = Storage(self)
        self.errors = []

        self.collect_factories_data()

    def _set_project_dir(self, absolute_path: str):
        self.project_dir = absolute_path.replace(f'{self.base_path}/', '').split('/')[0].replace('-', '_')

    def _get_project_dir_by_factory_name(self, factory_name):
        """Определяет директорию проекта, к которой относится фабрика"""

        # Фабрика с наименованием ActualReportStorageFactory есть и в core и в salary, при этом фабрика из salary
        # используется только в старых реализациях, соответственно используем только из core.
        excluded_factories = {
            'ActualReportStorageFactory',
        }

        if (
            factory_name not in excluded_factories and
            factory_name in self.records_file_paths_for_factories.records_in_project
        ):
            project_dir = self.project_dir
        else:
            if (
                self.project_dir in ('web_bb_vehicle', 'web_bb_food') and
                factory_name in self.records_file_paths_for_factories.storage['web_bb_accounting']
            ):
                project_dir = 'web_bb_accounting'
            else:
                project_dir = 'web_bb_core'

        return project_dir

    def collect_factories_records_data(self, records_filepath):
        with open(records_filepath) as f:
            ast_tree = ast.parse(f.read())

        for item in ast_tree.body:
            if isinstance(item, ast.FunctionDef):
                class_call = item.body[0].value
                corrected_keywords = set()

                for k in class_call.keywords:
                    if 'kwargs[' in ast.unparse(k.value):
                        unparsed_val = f"{k.value.value.slice.value}.pk"  # ".pk" очень важно!
                    else:
                        unparsed_val = ast.unparse(k.value)

                    corrected_keywords.add((k.arg, unparsed_val))

                key_id = frozenset(corrected_keywords)

                func_name = item.name

                self.all_func_names.add(func_name)

                factory_name = class_call.func.id

                try:
                    self.factories_records.records_in_project[factory_name][key_id] = (func_name, False)
                except KeyError:
                    self.factories_records.records_in_project[factory_name] = {key_id: (func_name, False)}

    def collect_factories_data(self):
        for dir_name, sub_dirs, file_list in os.walk(self.base_path):
            self._set_project_dir(dir_name)
            if 'pycache' in dir_name:
                continue
            elif 'features/factories' in dir_name:
                for f_name in file_list:
                    if '__init__.py' in f_name:
                        continue
                    elif f_name == 'records.py':
                        self.collect_factories_records_data(f'{dir_name}/{f_name}')
                    else:
                        with open(f'{dir_name}/{f_name}') as f:
                            ast_tree = ast.parse(f.read())

                        for item in ast_tree.body:
                            if isinstance(item, ast.ClassDef):
                                self.records_file_paths_for_factories.records_in_project[item.name] = f'{dir_name}/records.py'

    def process_ast_imports(self, ast_tree, app_request_strs):

        def exclusion_required(string):
            """Условие удаление импорта."""

            return (
                ('datetime.' not in app_request_strs and 'import datetime' in string) or
                'AppRequest,' in string or
                'Decimal,' in string or
                'ReportType,' in string
            )

        new_import_lines = []

        for item in ast_tree.body:
            if isinstance(item, ast.Import):
                new_import_lines.append(f'{ast.unparse(item)}\n')
            elif isinstance(item, ast.ImportFrom):
                if 'features.factories.records' in item.module:
                    new_import_lines.append(f'{ast.unparse(item)}\n')
                elif 'features.factories' in item.module:
                    import_module = item.module.split('features.factories')[0]
                    new_import_lines.append(f'from {import_module}features.factories.records import *\n')
                    for name_obj in item.names:
                        factory_name = name_obj.name
                        project_dir = self._get_project_dir_by_factory_name(factory_name)
                        self.factories_imports.storage[project_dir][factory_name] = item.module
                else:
                    import_names = ''.join(f'    {n.name},\n' for n in item.names)
                    new_import_lines.append(
                        f'from {item.module} import (\n{import_names})\n'
                    )

        return [i for i in new_import_lines if not exclusion_required(i)]

    def process_factories(self, ast_tree):
        new_function_strs = []

        for item in ast_tree.body:
            if isinstance(item, (ast.Import, ast.ImportFrom)):
                continue
            elif isinstance(item, ast.FunctionDef):
                if item.name == 'request_loader':
                    new_function_strs.append(f'\n\ndef {item.name}({ast.unparse(item.args)}):\n')
                    for el in item.body:
                        if re.search(r'.*Factory(___[\d\w]+)?\(', ast.unparse(el)):
                            factory_func_name = None
                            line_variables = ''

                            if isinstance(el, ast.Assign) or isinstance(el, ast.Expr):
                                variable_name = None

                                if isinstance(el, ast.Assign):
                                    line_variables = ' = '.join([t.id for t in el.targets])
                                    line_variables = f'{line_variables} = '
                                    variable_name = el.targets[0].id

                                    line_variables = re.sub(r'___[a-z\d]+', '', line_variables)
                                    variable_name = re.sub(r'___[a-z\d]+', '', variable_name)

                                keywords = set()
                                for k in el.value.keywords:
                                    keyword_value = ast.unparse(k.value)
                                    if keyword_value.endswith('.pk') and '___' in keyword_value:
                                        keyword_value = re.sub(r'___[a-z\d]+', '', keyword_value)
                                    keywords.add((k.arg, keyword_value))

                                factory_func_name = self.get_factory_func_name(
                                    el.value.func.id,
                                    keywords,
                                    variable_name=variable_name,
                                )

                            if factory_func_name:
                                new_function_strs.append(f'    {line_variables}{factory_func_name}(**locals())\n')
                        elif 'AppRequest' in ast.unparse(el):
                            break
                        else:
                            new_function_strs.append(f'    {ast.unparse(el)}\n')

        return new_function_strs

    def get_factory_func_name(self, factory_name, keywords, variable_name=None):
        new_func_name = "???!"

        factory_name, *_ = factory_name.split('___')
        project_dir = self._get_project_dir_by_factory_name(factory_name)

        key_id = frozenset(keywords)
        try:
            # exists_func_name, is_new = self.factories_records[factory_name].get(key_id, (None, False))
            exists_func_name, is_new = self.factories_records.storage[project_dir][factory_name].get(key_id, (None, False))
        except KeyError:
            exists_func_name = None
            # self.factories_records[factory_name] = {}
            self.factories_records.storage[project_dir][factory_name] = {}

        if not exists_func_name:
            if variable_name:
                new_func_name = f'get_{variable_name}'
            else:
                new_func_name = f'get_{factory_name.lower()}_{uuid.uuid4().hex[:6]}'

            if new_func_name in self.all_func_names:
                new_func_name = f'{new_func_name}_{uuid.uuid4().hex[:6]}'

            self.all_func_names.add(new_func_name)
            self.factories_records.storage[project_dir][factory_name][key_id] = (new_func_name, True)

        if exists_func_name:
            factory_func_name = exists_func_name
        else:
            factory_func_name = new_func_name

        return factory_func_name

    def process_file(self, file_path):
        if not ('/request_' in str(file_path) and str(file_path).endswith('.py')):
            return

        print(f'process {file_path}')

        new_file_lines = []

        with open(file_path) as f:
            content = f.read()

        if '.records import *' in content:
            print('already processed!')
            return

        ast_tree = ast.parse(content)
        _, response_param_name, _, app_request_strs = re.split(
            r'(response.*) = (context\.AppRequest|AppRequest)',
            content,
            maxsplit=1,
        )

        new_file_lines.append(
            ''.join(
                self.process_ast_imports(ast_tree, app_request_strs) +
                self.process_factories(ast_tree) +
                [f'\n    {response_param_name} = context.AppRequest{app_request_strs}']
            )
        )

        with open(file_path, 'w') as f:
            f.write(f'{"".join(new_file_lines).strip()}\n')

    def make_requests_factories_reuse(self, path):
        """
        """
        for path in path.split(','):
            self._set_project_dir(path)
            path = Path(path)

            if path.is_dir():
                for dir_name, sub_dirs, file_list in os.walk(path):
                    if 'pycache' in dir_name:
                        continue
                    if 'features/' in dir_name:
                        for file_name in file_list:
                            try:
                                self.process_file(f'{dir_name}/{file_name}')
                            except Exception as e:
                                self.errors.append((f'{dir_name}/{file_name}', str(e)))

            elif path.is_file():
                try:
                    self.process_file(path)
                except Exception as e:
                    self.errors.append((path, str(e)))

        self.update_factories_records()

        if self.errors:
            print('=' * 100)
            for path, err in self.errors:
                print(f'{path} - {err}')

    def update_factories_records(self):
        print('update_factories_records:')
        files_to_update = defaultdict(list)

        for project_dir, factories_records in self.factories_records.storage.items():
            for factory_name, records_data in factories_records.items():
                for keywords, (func_name, is_new) in records_data.items():
                    if is_new:
                        files_to_update[self.records_file_paths_for_factories.storage[project_dir][factory_name]].append(
                            (func_name, factory_name, keywords)
                        )

        for file_path, factory_functions in files_to_update.items():
            import_decimal = False
            import_datetime = False

            file_lines = []
            factory_import_lines = set()

            for func_name, factory_name, keywords in factory_functions:
                sorted_keywords = []

                for k, v in sorted(keywords):
                    if v.endswith('.pk'):
                        kwarg_var = v.split('.')[0]
                        v = f'kwargs[\'{kwarg_var}\'].pk'
                    elif 'Decimal' in v:
                        import_decimal = True
                    elif 'datetime' in v:
                        import_datetime = True

                    sorted_keywords.append(f'        {k}={v},')

                sorted_keywords = '\n'.join(sorted_keywords)
                file_lines.append(
                    f'def {func_name}(**kwargs):\n'
                    f'    return {factory_name}(\n'
                    f'{sorted_keywords}\n'
                    f'    )\n\n'
                )
                # self.factories_records[factory_name][keywords] = (func_name, False)
                project_dir = self._get_project_dir_by_factory_name(factory_name)
                self.factories_records.storage[project_dir][factory_name][keywords] = (func_name, False)
                factory_import_lines.add(f'from {self.factories_imports.storage[project_dir][factory_name]} import *')

            imports_lines = []

            if os.path.isfile(file_path):
                print(f'\tupdate: {file_path}')

                with open(file_path, 'r+') as f:
                    exists_file_content = f.read()

                    if import_decimal and 'Decimal' not in exists_file_content:
                        imports_lines.append('from decimal import Decimal')
                    if import_datetime and 'datetime' not in exists_file_content:
                        imports_lines.append('import datetime')

                    f.seek(0)
                    f.write('\n'.join(
                        imports_lines + [exists_file_content, ''] + file_lines
                    ))
                    f.truncate()
            else:
                print(f'\tcreate: {file_path}')

                if import_decimal and 'Decimal':
                    imports_lines.append('from decimal import Decimal')
                if import_datetime and 'datetime':
                    imports_lines.append('import datetime')

                imports_lines.extend(list(factory_import_lines) + ['\n'])

                with open(file_path, 'w+') as f:
                    f.write('\n'.join(
                        imports_lines + file_lines
                    ))


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Перенос и объединение кода получения фабрик.')

    parser.add_argument('base_path', help='Базовая директория расположения web_bb репозиториев', type=str)
    parser.add_argument('path', help='Путь к файлу или директории для которой нужно выполнить обработку', type=str)

    args = parser.parse_args()

    reuse_proc = FactoriesReuseProcessor(args.base_path)
    reuse_proc.make_requests_factories_reuse(args.path)
